/*

	jackmixdesk.c
	Copyright (C) 2005,2007  Nicholas J. Humfrey, Ulrich Lorenz Schlueter

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <strings.h>
#include <signal.h>
#include <math.h>
#include <ctype.h>
#include <getopt.h>
#include <pthread.h>
#include <time.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>

#include <jack/jack.h>

#include <lash/lash.h>
#include <lo/lo.h>

#include <libxml/xmlwriter.h>
#include <libxml/parser.h>
#include <libxml/encoding.h>

#include "config.h"
#include "db.h"

#define		DEFAULT_CLIENT_NAME		"JackMixDesk"	/* Default name of client */
#define		DEFAULT_OSC_PORT		"16000"			/* Default port of osc server */
#define		DEFAULT_LASH_PORT		"14541"			/* Default port of lashd */
#define		DEFAULT_CHANNEL_COUNT	(4)				/* Default number of input channels */
#define		DEFAULT_PREFADER_COUNT	(1)				/* Default number of prefader sends defaults to one (master) */
#define		DEFAULT_POSTFADER_COUNT	(0)				/* Default number of postfader sends */
#define		DEFAULT_LABEL_LEN		(32)			/* Max length of channel label strings */
#define		GAIN_FADE_SLEEP			(10000)			/* µs sleep between fade steps */
#define 	DEFAULT_CONFIG_FILE		".mixdeskrc"	/* Default config file name */
#define 	DEFAULT_CONFIG_PATH		"."				/* Default config file path */
#define 	DEFAULT_XML_ENCODING	"UTF-8"			/* Default xml encoding */


typedef struct {
	float current_gain;				/* decibels represents an amp */
	float desired_gain;				/* decibels represents an amp  */
	float current_gain_l;			/* decibels represents an amp  */
	float current_gain_r;			/* decibels represents an amp  */
	float desired_gain_l;			/* decibels represents an amp */
	float desired_gain_r;			/* decibels represents an amp */
	float mix_gain_r;				/* decibels product of all gains */
	float mix_gain_l;				/* decibels product of all gains */
} jmd_send_t;

typedef struct {
	char label[DEFAULT_LABEL_LEN];	/* Label for channel */
	float mute;
	int stereo;						/* make channel stereo, bind to next channel */
	char from_port[64];				/* first port in_port is connected to */
	char **from_ports;				/* all ports in_port is connected to */
	jack_port_t *in_port;			/* Mono input */
	jmd_send_t *send;				/* Aux sends */
} jmd_channel_t;

typedef struct {
	char label[DEFAULT_LABEL_LEN];	/* Label for output and sends*/
	float current_gain;				/* decibels represents an amp */
	float desired_gain;				/* decibels represents an amp */
	float mute;
	int link;				/* Gain linked with next channel */
	char to_port_l[64];				/* port where left port will be connected to */
	char to_port_r[64];
	char **to_ports_l;				/* all ports the left port is connected to jack_port_get_connections() */
	char **to_ports_r;
	jack_port_t *out_port_l;		/* Left output port */
	jack_port_t *out_port_r;		/* Right output port */
} jmd_output_t;

typedef struct {
	unsigned int channel_count;
	unsigned int prefader_count;
	unsigned int postfader_count;
	unsigned int autoconnect;
	char client_name[DEFAULT_LABEL_LEN];
	char osc_mixer_port[DEFAULT_LABEL_LEN];
	char xml_config_file[DEFAULT_LABEL_LEN];
	char xml_config_path[256];
	char lash_project[DEFAULT_LABEL_LEN];
	char lash_server[DEFAULT_LABEL_LEN];
	char lash_id[64];
	jack_client_t *jack_client;
	lash_client_t *lash_client;
	jmd_channel_t *channel;
	jmd_output_t *output;
	lo_server_thread st;
	lo_server serv;
	xmlDocPtr doc;
} jmd_mixer_t;

static struct option opts_long[] ={
		/* These options set a flag. */
		{"verbose", no_argument,       0, 'v'},
		{"quiet",   no_argument,       0, 'q'},
		/* These options don't set a flag.
		 * We distinguish them by their indices. */
		{"autoconnect",	no_argument, 		0, 'a'},
		{"name",  		required_argument,	0, 'n'},
		{"channels",  	required_argument,	0, 'c'},
		{"pre",			required_argument,	0, 'e'},
		{"post",		required_argument,	0, 'o'},
		{"port",		required_argument,	0, 'm'},
		{"file",    	required_argument,	0, 'f'},
		{"path",    	required_argument,	0, 'p'},
		{"lash-project",required_argument,	0, 'l'},
		{"lash-server",	required_argument,	0, 's'},
		{"lash-id",		required_argument,	0, 'i'},
		{0, 0, 0, 0}
};

unsigned int verbose=0;
unsigned int quiet=0;
unsigned int running=1;

static int xml_parse_mixer_send (xmlNodePtr cur, jmd_mixer_t *mixer, int c){

	xmlChar *key;
	char tmp[12];
	int s;
	float g;
	cur = cur->xmlChildrenNode;
	while (cur != NULL) {
	    if ((!xmlStrcmp(cur->name, (const xmlChar *)"ID"))) {
		    key = xmlNodeListGetString(mixer->doc, cur->xmlChildrenNode, 1);
		    if (verbose==2) printf("SEND\n");
		    sprintf(tmp,"%s",key);
		    s=atoi(tmp);
		    if (verbose==2) printf("ID: %i\n", s);
		    xmlFree(key);
 	    } else if ((!xmlStrcmp(cur->name, (const xmlChar *)"GAIN"))) {
		    key = xmlNodeListGetString(mixer->doc, cur->xmlChildrenNode, 1);
		    sprintf(tmp,"%s",key);
		    g=atof(tmp);
		    if (verbose==2) printf("GAIN: %f\n", g);
		    mixer->channel[c].send[s-1].current_gain=g;
		    mixer->channel[c].send[s-1].desired_gain=g;
		    xmlFree(key);
 	    } else if ((!xmlStrcmp(cur->name, (const xmlChar *)"GAIN_L"))) {
		    key = xmlNodeListGetString(mixer->doc, cur->xmlChildrenNode, 1);
		    sprintf(tmp,"%s",key);
		    g=atof(tmp);
		    if (verbose==2) printf("GAIN_L: %f\n", g);
		    mixer->channel[c].send[s-1].current_gain_l=g;
		    mixer->channel[c].send[s-1].desired_gain_l=g;
		    xmlFree(key);
 	    } else if ((!xmlStrcmp(cur->name, (const xmlChar *)"GAIN_R"))) {
		    key = xmlNodeListGetString(mixer->doc, cur->xmlChildrenNode, 1);
		    sprintf(tmp,"%s",key);
		    g=atof(tmp);
		    if (verbose==2) printf("GAIN_R: %f\n", g);
		    mixer->channel[c].send[s-1].current_gain_r=g;
		    mixer->channel[c].send[s-1].desired_gain_r=g;
		    xmlFree(key);
 	    }
	cur = cur->next;
	}
    return 0;
}

static int xml_parse_mixer_channel (xmlNodePtr cur, jmd_mixer_t *mixer){

	xmlChar *key=NULL;
	char tmp[32];
	int c,s;
	float g;
	cur = cur->xmlChildrenNode;
	while (cur != NULL) {
	    if ((!xmlStrcmp(cur->name, (const xmlChar *)"ID"))) {
		    key = xmlNodeListGetString(mixer->doc, cur->xmlChildrenNode, 1);
		    if (verbose==2) printf("INPUT\n");
		    sprintf(tmp,"%s",key);
		    c=atoi(tmp);
		    if (verbose==2) printf("ID: %i\n", c);
		    xmlFree(key);
 	    } else if ((!xmlStrcmp(cur->name, (const xmlChar *)"PORT_NAME"))) {
		    key = xmlNodeListGetString(mixer->doc, cur->xmlChildrenNode, 1);
		    if (verbose==2) printf("PORT_NAME: %s\n", key);
		    sprintf(mixer->channel[c-1].label,"%s",key);
		    xmlFree(key);
 	    }  else if ((!xmlStrcmp(cur->name, (const xmlChar *)"PORT_FROM"))) {
		    key = xmlNodeListGetString(mixer->doc, cur->xmlChildrenNode, 1);
		    if (verbose==2) printf("PORT_FROM: %s\n", key);
		    sprintf(mixer->channel[c-1].from_port,"%s",key);
		    xmlFree(key);
 	    } else if ((!xmlStrcmp(cur->name, (const xmlChar *)"MUTE"))) {
		    key = xmlNodeListGetString(mixer->doc, cur->xmlChildrenNode, 1);
		    sprintf(tmp,"%s",key);
		    g=atof(tmp);
		    mixer->channel[c-1].mute=g;
		    if (verbose==2) printf("MUTE: %f\n",g);
		    xmlFree(key);
 	    } else  if ((!xmlStrcmp(cur->name, (const xmlChar *)"STEREO"))){
		    key = xmlNodeListGetString(mixer->doc, cur->xmlChildrenNode, 1);
		    sprintf(tmp,"%s",key);
		    s=atoi(tmp);
		    mixer->channel[c-1].stereo=s;
		    if (verbose==2) printf("STEREO: %i\n", s);
		    xmlFree(key);
 	    } else if ((!xmlStrcmp(cur->name, (const xmlChar *)"SEND"))) {
			xml_parse_mixer_send(cur, mixer, c-1);
 	    }
	cur = cur->next;
	}
    return 0;
}

static int xml_parse_mixer_output (xmlNodePtr cur, jmd_mixer_t *mixer){

	xmlChar *key;
	char tmp[12];
	int o,l;
	float g;

	cur = cur->xmlChildrenNode;
	while (cur != NULL) {
	    if ((!xmlStrcmp(cur->name, (const xmlChar *)"ID"))) {
		    key = xmlNodeListGetString(mixer->doc, cur->xmlChildrenNode, 1);
		    if (verbose==2) printf("OUTPUT\n");
		    sprintf(tmp,"%s",key);
		    o=atoi(tmp);
		    if (verbose==2) printf("ID: %i\n", o);
		    xmlFree(key);
 	    } else if ((!xmlStrcmp(cur->name, (const xmlChar *)"PORT_NAME"))) {
		    key = xmlNodeListGetString(mixer->doc, cur->xmlChildrenNode, 1);
		    if (verbose==2) printf("PORT_NAME: %s\n", key);
		    sprintf(mixer->output[o-1].label,"%s",key);
		    xmlFree(key);
 	    } else if ((!xmlStrcmp(cur->name, (const xmlChar *)"PORT_TO_L"))) {
		    key = xmlNodeListGetString(mixer->doc, cur->xmlChildrenNode, 1);
		    if (verbose==2) printf("PORT_TO_L: %s\n", key);
		    sprintf(mixer->output[o-1].to_port_l,"%s",key);
		    xmlFree(key);
 	    } else if ((!xmlStrcmp(cur->name, (const xmlChar *)"PORT_TO_R"))) {
		    key = xmlNodeListGetString(mixer->doc, cur->xmlChildrenNode, 1);
		    if (verbose==2) printf("PORT_TO_R: %s\n", key);
		    sprintf(mixer->output[o-1].to_port_r,"%s",key);
		    xmlFree(key);
 	    } else if ((!xmlStrcmp(cur->name, (const xmlChar *)"GAIN"))){
		    key = xmlNodeListGetString(mixer->doc, cur->xmlChildrenNode, 1);
		    sprintf(tmp,"%s",key);
		    g=atof(tmp);
		    if (verbose==2) printf("GAIN: %f\n", g);
		    mixer->output[o-1].current_gain=g;
		    mixer->output[o-1].desired_gain=g;
		    xmlFree(key);
 	    } else if ((!xmlStrcmp(cur->name, (const xmlChar *)"MUTE"))){
		    key = xmlNodeListGetString(mixer->doc, cur->xmlChildrenNode, 1);
		    sprintf(tmp,"%s",key);
		    g=atof(tmp);
		    if (verbose==2) printf("MUTE: %f\n", g);
		    mixer->output[o-1].mute=g;
		    xmlFree(key);
 	    } else if ((!xmlStrcmp(cur->name, (const xmlChar *)"LINK"))){
		    key = xmlNodeListGetString(mixer->doc, cur->xmlChildrenNode, 1);
		    sprintf(tmp,"%s",key);
		    l=atoi(tmp);
		    if (verbose==2) printf("LINK: %i\n", l);
		    mixer->output[o-1].link=l;
		    xmlFree(key);
 	    }
	cur = cur->next;
	}
    return 0;
}

static int xml_parse_base_config(jmd_mixer_t *mixer){

	xmlNodePtr cur;
	xmlChar *key;
	char config_file[256];
	char tmp[12];
	int c;

	sprintf(config_file,"%s/%s",mixer->xml_config_path,mixer->xml_config_file);
	mixer->doc = xmlParseFile(config_file);

	if (mixer->doc == NULL ) {
		fprintf(stderr,"Document not parsed successfully. \n");
		return 1;
	}

	cur = xmlDocGetRootElement(mixer->doc);

	if (cur == NULL) {
		fprintf(stderr,"empty document\n");
		xmlFreeDoc(mixer->doc);
		return 1;
	}

	if (xmlStrcmp(cur->name, (const xmlChar *) "MIXER")) {
		fprintf(stderr,"document of the wrong type, root node != MIXER");
		xmlFreeDoc(mixer->doc);
		return 1;
	}

	cur = cur->xmlChildrenNode;
	while (cur != NULL) {
		if ((!xmlStrcmp(cur->name, (const xmlChar *)"CLIENT_NAME"))){
			key = xmlNodeListGetString(mixer->doc, cur->xmlChildrenNode, 1);
			if (verbose==2) printf("CLIENT_NAME %s\n",key);
			sprintf(mixer->client_name,"%s",key);
			xmlFree(key);
		} else if ((!xmlStrcmp(cur->name, (const xmlChar *)"CHANNEL_COUNT"))){
			if (verbose==2) printf("CHANNEL_COUNT %s\n",xmlNodeListGetString(mixer->doc, cur->xmlChildrenNode, 1));
			key = xmlNodeListGetString(mixer->doc, cur->xmlChildrenNode, 1);
		    sprintf(tmp,"%s",key);
		    c=atoi(tmp);
			mixer->channel_count=c;
		} else if ((!xmlStrcmp(cur->name, (const xmlChar *)"PREFADER_COUNT"))){
			if (verbose==2) printf("PREFADER_COUNT %s\n",xmlNodeListGetString(mixer->doc, cur->xmlChildrenNode, 1));
			key = xmlNodeListGetString(mixer->doc, cur->xmlChildrenNode, 1);
		    sprintf(tmp,"%s",key);
		    c=atoi(tmp);
			mixer->prefader_count=c;
		} else if ((!xmlStrcmp(cur->name, (const xmlChar *)"POSTFADER_COUNT"))){
			if (verbose==2) printf("POSTFADER_COUNT %s\n",xmlNodeListGetString(mixer->doc, cur->xmlChildrenNode, 1));
			key = xmlNodeListGetString(mixer->doc, cur->xmlChildrenNode, 1);
		    sprintf(tmp,"%s",key);
		    c=atoi(tmp);
			mixer->postfader_count=c;
		} else if ((!xmlStrcmp(cur->name, (const xmlChar *)"OSC_PORT"))){
			key = xmlNodeListGetString(mixer->doc, cur->xmlChildrenNode, 1);
			if (verbose==2) printf("OSC_PORT %s\n",key);
			sprintf(mixer->osc_mixer_port,"%s",key);
			xmlFree(key);
		}
	cur = cur->next;
	}
	return 0;
}

static int xml_parse_io_config(jmd_mixer_t *mixer){

	xmlNodePtr cur;

	if (mixer->doc == NULL ) {
		fprintf(stderr,"Document not parsed successfully. \n");
		return 1;
	}

	cur = xmlDocGetRootElement(mixer->doc);

	if (cur == NULL) {
		fprintf(stderr,"empty document\n");
		xmlFreeDoc(mixer->doc);
		return 1;
	}

	if (xmlStrcmp(cur->name, (const xmlChar *) "MIXER")) {
		fprintf(stderr,"document of the wrong type, root node != MIXER");
		xmlFreeDoc(mixer->doc);
		return 1;
	}

	cur = cur->xmlChildrenNode;
	while (cur != NULL) {
		if ((!xmlStrcmp(cur->name, (const xmlChar *)"CHANNEL"))){
			xml_parse_mixer_channel(cur, mixer);
		} else {
			xml_parse_mixer_output(cur, mixer);
		}
	cur = cur->next;
	}
	return 0;
}

static xmlChar *xml_convert_input(const char *in, const char *encoding)
{
    xmlChar *out;
    int r,s,o,t;
    xmlCharEncodingHandlerPtr handler;

    if (in == 0)
        return 0;

    handler = xmlFindCharEncodingHandler(encoding);

    if (!handler) {
        printf("ConvertInput: no encoding handler found for '%s'\n",
               encoding ? encoding : "");
        return 0;
    }

    s = (int) strlen(in) + 1;
    o = s * 2 - 1;
    out = (unsigned char *) xmlMalloc((size_t) o);

    if (out != 0) {
        t = s - 1;
        r = handler->input(out, &o, (const xmlChar *) in, &t);
        if ((r < 0) || (t - s + 1)) {
            if (r < 0) {
                printf("ConvertInput: conversion wasn't successful.\n");
            } else {
                printf
                    ("ConvertInput: conversion wasn't successful. converted: %i octets.\n",
                     t);
            }

            xmlFree(out);
            out = 0;
        } else {
            out = (unsigned char *) xmlRealloc(out, o + 1);
            out[o] = 0;  /*null terminating out */
        }
    } else {
        printf("ConvertInput: no mem\n");
    }
    return out;
}

static int xml_write_config(jmd_mixer_t *mixer)
{
    int rc,c,s;
    xmlTextWriterPtr writer;
    xmlChar *tmp;
    char config_file[256];

	sprintf(config_file,"%s/%s",mixer->xml_config_path,mixer->xml_config_file);
    /* Create a new XmlWriter for xml_config_file, with no compression. */
    writer = xmlNewTextWriterFilename(config_file, 0);
    if (writer == NULL) {
        printf("xml_write_config: Error creating the xml writer\n");
        return 1;
    }

    /* Start the document with the xml default for the version,
     * encoding UTF-8 and the default for the standalone
     * declaration. */
    rc = xmlTextWriterStartDocument(writer, NULL, DEFAULT_XML_ENCODING, NULL);
    if (rc < 0) {
        printf
            ("xml_write_config: Error at xmlTextWriterStartDocument\n");
        return 1;
    }

    /* Start an element named MIXER. Since thist is the first
     * element, this will be the root element of the document. */
    rc = xmlTextWriterStartElement(writer, BAD_CAST "MIXER");
    if (rc < 0) {
        printf
            ("xml_write_config: Error at xmlTextWriterStartElement\n");
        return 1;
    }

    /* Add an attribute with name "version" and value "1.0" to MIXER. */
    rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "version",
                                     BAD_CAST "1.0");
    if (rc < 0) {
        printf
            ("xml_write_config: Error at xmlTextWriterWriteAttribute\n");
        return 1;
    }

    /* Start an element named HEADER. Since thist is the first
     * element, this will be the root element of the document. */
    rc = xmlTextWriterStartElement(writer, BAD_CAST "HEADER");
    if (rc < 0) {
        printf
            ("xml_write_config: Error at xmlTextWriterStartElement\n");
        return 1;
    }

    /* Write a comment as child of HEADER.
     * Please observe, that the input to the xmlTextWriter functions
     * HAS to be in UTF-8, even if the output XML is encoded
     * in iso-8859-1 */
    tmp = xml_convert_input("This is version 1.0 of the JackMixDesk config file.",
                       DEFAULT_XML_ENCODING);
    rc = xmlTextWriterWriteComment(writer, tmp);
    if (rc < 0) {
        printf
            ("xml_write_config: Error at xmlTextWriterWriteComment\n");
        return 1;
    }
    if (tmp != NULL) xmlFree(tmp);

    /* Write a comment as child of HEADER. */
    tmp = xml_convert_input("Copyright (C) 2007 Ulrich Lorenz Schlueter <audio-mobster@gmx.de>",
                       DEFAULT_XML_ENCODING);
    rc = xmlTextWriterWriteComment(writer, tmp);
    if (rc < 0) {
        printf
            ("xml_write_config: Error at xmlTextWriterWriteComment\n");
        return 1;
    }
    if (tmp != NULL) xmlFree(tmp);

    /* Close the element named HEADER. */
    rc = xmlTextWriterEndElement(writer);
    if (rc < 0) {
        printf
            ("xml_write_config: Error at xmlTextWriterEndElement\n");
        return 1;
    }

    /* Write an element named CLIENT_NAME as child of MIXER. */
    rc = xmlTextWriterWriteFormatElement(writer, BAD_CAST "CLIENT_NAME","%s", mixer->client_name);
    if (rc < 0) {
        printf
            ("xml_write_config: Error at xmlTextWriterWriteFormatElement\n");
        return 1;
    }

	/* Write an element named CHANNEL_COUNT as child of MIXER. */
    rc = xmlTextWriterWriteFormatElement(writer, BAD_CAST "CHANNEL_COUNT","%i", mixer->channel_count);
    if (rc < 0) {
        printf
            ("xml_write_config: Error at xmlTextWriterWriteFormatElement\n");
        return 1;
    }

    /* Write an element named PREFADER_COUNT as child of MIXER. */
    rc = xmlTextWriterWriteFormatElement(writer, BAD_CAST "PREFADER_COUNT","%i", mixer->prefader_count);
    if (rc < 0) {
        printf
            ("xml_write_config: Error at xmlTextWriterWriteFormatElement\n");
        return 1;
    }

    /* Write an element named POSTFADER_COUNT as child of MIXER. */
    rc = xmlTextWriterWriteFormatElement(writer, BAD_CAST "POSTFADER_COUNT","%i", mixer->postfader_count);
    if (rc < 0) {
        printf
            ("xml_write_config: Error at xmlTextWriterWriteFormatElement\n");
        return 1;
    }

    /* Write an element named OSC_PORT as child of MIXER. */
    rc = xmlTextWriterWriteFormatElement(writer, BAD_CAST "OSC_PORT","%s", mixer->osc_mixer_port);
    if (rc < 0) {
        printf
            ("xml_write_config: Error at xmlTextWriterWriteFormatElement\n");
        return 1;
    }

    for (c=0;c<mixer->channel_count;c++){
	    /* Start an element named CHANNEL as child of MIXER. */
	    rc = xmlTextWriterStartElement(writer, BAD_CAST "CHANNEL");
	    if (rc < 0) {
	        printf
	            ("xml_write_config: Error at xmlTextWriterStartElement\n");
	        return 1;
	    }

	    /* Write an element named ID as child of CHANNEL. */
	    rc = xmlTextWriterWriteFormatElement(writer, BAD_CAST "ID","%d", c+1);
	    if (rc < 0) {
	        printf
	            ("xml_write_config: Error at xmlTextWriterWriteFormatElement\n");
	        return 1;
	    }

	    /* Write an element named PORT_NAME as child of CHANNEL. */
	    tmp = xml_convert_input(mixer->channel[c].label, DEFAULT_XML_ENCODING);
	    rc = xmlTextWriterWriteElement(writer, BAD_CAST "PORT_NAME", tmp);
	    if (rc < 0) {
	        printf
	            ("testXmlwriterFilename: Error at xmlTextWriterWriteElement\n");
	        return 1;
	    }
	    if (tmp != NULL) xmlFree(tmp);

	    /* Write an element named PORT_FROM as child of CHANNEL. */
	    if (mixer->channel[c].from_ports==NULL){
	    	tmp = xml_convert_input("none", DEFAULT_XML_ENCODING);
			rc = xmlTextWriterWriteElement(writer, BAD_CAST "PORT_FROM", tmp);
	    } else {
	    	tmp = xml_convert_input(mixer->channel[c].from_ports[0], DEFAULT_XML_ENCODING);
	    	rc = xmlTextWriterWriteElement(writer, BAD_CAST "PORT_FROM", tmp);
	    }

	    if (rc < 0) {
	        printf
	            ("testXmlwriterFilename: Error at xmlTextWriterWriteElement\n");
	        return 1;
	    }

	    if (tmp != NULL) xmlFree(tmp);

	    /* Write an element named MUTE as child of CHANNEL. */
	    rc = xmlTextWriterWriteFormatElement(writer, BAD_CAST "MUTE","%f", mixer->channel[c].mute);
	    if (rc < 0) {
	        printf
	            ("xml_write_config: Error at xmlTextWriterWriteFormatElement\n");
	        return 1;
	    }

	    /* Write an element named STEREO as child of CHANNEL. */
	    rc = xmlTextWriterWriteFormatElement(writer, BAD_CAST "STEREO","%i", mixer->channel[c].stereo);
	    if (rc < 0) {
	        printf
	            ("xml_write_config: Error at xmlTextWriterWriteFormatElement\n");
	        return 1;
	    }

	    for(s=0;s<mixer->prefader_count+mixer->postfader_count;s++){
		    /* Start an element named SEND as child of CHANNEL. */
		    rc = xmlTextWriterStartElement(writer, BAD_CAST "SEND");
		    if (rc < 0) {
		        printf
		            ("xml_write_config: Error at xmlTextWriterStartElement\n");
		        return 1;
		    }

		    /* Write an element named ID as child of SEND. */
		    rc = xmlTextWriterWriteFormatElement(writer, BAD_CAST "ID","%d", s+1);
		    if (rc < 0) {
		        printf
		            ("xml_write_config: Error at xmlTextWriterWriteFormatElement\n");
		        return 1;
		    }

		    /* Write an element named "GAIN" as child of SEND. */
		    rc = xmlTextWriterWriteFormatElement(writer, BAD_CAST "GAIN","%f", mixer->channel[c].send[s].current_gain);
		    if (rc < 0) {
		        printf
		            ("xml_write_config: Error at xmlTextWriterWriteFormatElement\n");
		        return 1;
		    }

		    /* Write an element named "GAIN_L" as child of SEND. */
		    rc = xmlTextWriterWriteFormatElement(writer, BAD_CAST "GAIN_L","%f", mixer->channel[c].send[s].current_gain_l);
		    if (rc < 0) {
		        printf
		            ("xml_write_config: Error at xmlTextWriterWriteFormatElement\n");
		        return 1;
		    }

		    /* Write an element named "GAIN_R" as child of SEND. */
		    rc = xmlTextWriterWriteFormatElement(writer, BAD_CAST "GAIN_R","%f", mixer->channel[c].send[s].current_gain_r);
		    if (rc < 0) {
		        printf
		            ("xml_write_config: Error at xmlTextWriterWriteFormatElement\n");
		        return 1;
		    }

		    /* Close the element named SEND. */
		    rc = xmlTextWriterEndElement(writer);
		    if (rc < 0) {
		        printf
		            ("xml_write_config: Error at xmlTextWriterEndElement\n");
		        return 1;
		    }
	    }
	    /* Close the element named CHANNEL. */
	    rc = xmlTextWriterEndElement(writer);
	    if (rc < 0) {
	        printf
	            ("xml_write_config: Error at xmlTextWriterEndElement\n");
	        return 1;
	    }
    }

    for(s=0;s<mixer->prefader_count+mixer->postfader_count;s++){

	    /* Start an element named OUTPUT. */
	    rc = xmlTextWriterStartElement(writer, BAD_CAST "OUTPUT");
	    if (rc < 0) {
	        printf
	            ("xml_write_config: Error at xmlTextWriterStartElement\n");
	        return 1;
	    }

	    /* Write an element named ID as child of OUTPUT. */
	    rc = xmlTextWriterWriteFormatElement(writer, BAD_CAST "ID","%d", s+1);
	    if (rc < 0) {
	        printf
	            ("xml_write_config: Error at xmlTextWriterWriteFormatElement\n");
	        return 1;
	    }

	    /* Write an element named PORT_NAME as child of OUTPUT. */
	    tmp = xml_convert_input(mixer->output[s].label, DEFAULT_XML_ENCODING);
	    rc = xmlTextWriterWriteElement(writer, BAD_CAST "PORT_NAME", tmp);
	    if (rc < 0) {
	        printf
	            ("testXmlwriterFilename: Error at xmlTextWriterWriteElement\n");
	        return 1;
	    }
	    if (tmp != NULL) xmlFree(tmp);

	    /* Write an element named PORT_TO_L as child of OUTPUT.*/
	    if (mixer->output[s].to_ports_l==NULL){
	    	tmp = xml_convert_input("none", DEFAULT_XML_ENCODING);
			rc = xmlTextWriterWriteElement(writer, BAD_CAST "PORT_TO_L", tmp);
	    } else {
	    	tmp = xml_convert_input(mixer->output[s].to_ports_l[0], DEFAULT_XML_ENCODING);
	    	rc = xmlTextWriterWriteElement(writer, BAD_CAST "PORT_TO_L", tmp);
	    }
	    if (rc < 0) {
	        printf
	            ("testXmlwriterFilename: Error at xmlTextWriterWriteElement\n");
	        return 1;
	    }
	    if (tmp != NULL) xmlFree(tmp);

	    /* Write an element named PORT_TO_R as child of OUTPUT.*/
	    if (mixer->output[s].to_ports_r==NULL){
	    	tmp = xml_convert_input("none", DEFAULT_XML_ENCODING);
			rc = xmlTextWriterWriteElement(writer, BAD_CAST "PORT_TO_R", tmp);
	    } else {
	    	tmp = xml_convert_input(mixer->output[s].to_ports_r[0], DEFAULT_XML_ENCODING);
	   		rc = xmlTextWriterWriteElement(writer, BAD_CAST "PORT_TO_R", tmp);
	    }
    	if (rc < 0) {
        	printf
            ("testXmlwriterFilename: Error at xmlTextWriterWriteElement\n");
        	return 1;
    	}

	    if (tmp != NULL) xmlFree(tmp);

	    /* Write an element named "GAIN" as child of OUTPUT. */
	    rc = xmlTextWriterWriteFormatElement(writer, BAD_CAST "GAIN","%f", mixer->output[s].current_gain);
	    if (rc < 0) {
	        printf
	            ("xml_write_config: Error at xmlTextWriterWriteFormatElement\n");
	        return 1;
	    }

	    /* Write an element named MUTE as child of OUTPUT. */
	    rc = xmlTextWriterWriteFormatElement(writer, BAD_CAST "MUTE","%f", mixer->output[s].mute);
	    if (rc < 0) {
	        printf
	            ("xml_write_config: Error at xmlTextWriterWriteFormatElement\n");
	        return 1;
	    }

	    /* Write an element named LINK as child of OUTPUT. */
	    rc = xmlTextWriterWriteFormatElement(writer, BAD_CAST "LINK","%d", mixer->output[s].link);
	    if (rc < 0) {
	        printf
	            ("xml_write_config: Error at xmlTextWriterWriteFormatElement\n");
	        return 1;
	    }

	    /* Close the element named OUTPUT. */
	    rc = xmlTextWriterEndElement(writer);
	    if (rc < 0) {
	        printf
	            ("xml_write_config: Error at xmlTextWriterEndElement\n");
	        return 1;
	    }
    }

    /* Here we could close the element MIXER using the
     * function xmlTextWriterEndElement, but since we do not want to
     * write any other elements, we simply call xmlTextWriterEndDocument,
     * which will do all the work. */
    rc = xmlTextWriterEndDocument(writer);
    if (rc < 0) {
        printf
            ("xml_write_config: Error at xmlTextWriterEndDocument\n");
        return 1;
    }
    xmlFreeTextWriter(writer);
    return 0;
}

static int jack_process_audio(jack_nframes_t nframes, void *arg)
{
	int c,s;
	jmd_mixer_t *mixer = (jmd_mixer_t*)arg;
	jack_nframes_t n=0;

	/* Step through all outputs and mix each input into each output buffer */
	for (s=0;s<mixer->prefader_count+mixer->postfader_count;s++) {

        jack_default_audio_sample_t *out_left =
                jack_port_get_buffer(mixer->output[s].out_port_l, nframes);
        jack_default_audio_sample_t *out_right =
                jack_port_get_buffer(mixer->output[s].out_port_r, nframes);

        /* Put silence into the outputs */
        for ( n=0; n<nframes; n++ ) {
                out_left[ n ] = 0;
                out_right[ n ] = 0;
        }

        /* Mix each input into the output buffer
         * all mixing is done here */
        for (c=0;c<mixer->channel_count;c++) {

            jack_default_audio_sample_t *in =
                    jack_port_get_buffer(mixer->channel[c].in_port, nframes);

            for ( n=0; n<nframes; n++ ) {
                    out_left[ n ] += (in[ n ] * mixer->channel[c].send[s].mix_gain_l);
                    out_right[ n ] += (in[ n ] * mixer->channel[c].send[s].mix_gain_r);
            }
        }
	}
	return 0;
}

static void jack_shutdown_callback(void *arg)
{
	running = 0;
}

static int jack_init_client(jmd_mixer_t *mixer)
{
	jack_status_t status;

	printf("Connecting to Jack...\n");

	/* Register with Jack */
	if ((mixer->jack_client = jack_client_open(mixer->client_name, JackNullOption, &status)) == 0) {
		fprintf(stderr, "Failed to start jack client: %d\n", status);
		return 1;;
	}


	/* Register shutdown callback */
	jack_on_shutdown (mixer->jack_client, jack_shutdown_callback, mixer );

	/* Register the peak audio callback */
	jack_set_process_callback(mixer->jack_client, jack_process_audio, mixer);

	printf("Done\n");

	return 0;
}

static void jack_finish_client(jmd_mixer_t *mixer)
{
	/* Leave the Jack graph */
	jack_client_close(mixer->jack_client);

	if (verbose) printf( "Disconnected from Jack.\n");
}

jack_port_t* jack_create_in_port(const char* name, jmd_mixer_t *mixer)
{
	int port_name_size = jack_port_name_size();
	char *port_name = malloc( port_name_size );
	jack_port_t *port;

	snprintf( port_name, port_name_size, "%s", name );
	port = jack_port_register(mixer->jack_client, port_name, JACK_DEFAULT_AUDIO_TYPE, JackPortIsInput, 0);
	if (!port) {
		fprintf(stderr, "Cannot register input port '%s'.\n", port_name);
		exit(1);
	}
	return port;
}

jack_port_t* jack_create_out_port(const char* name, const char* side, jmd_mixer_t *mixer)
{
	int port_name_size = jack_port_name_size();
	char *port_name = malloc( port_name_size );
	jack_port_t *port;

	snprintf( port_name, port_name_size, "%s %s",name, side );
	port = jack_port_register(mixer->jack_client, port_name, JACK_DEFAULT_AUDIO_TYPE, JackPortIsOutput, 0);
	if (!port) {
		fprintf(stderr, "Cannot register output port '%s'.\n", port_name);
		exit(1);
	}
	return port;
}

static int jack_init_chan_ports(jmd_mixer_t *mixer)
{
	int ch;
	/* Initialise each of the channels */
	for(ch=0; ch<mixer->channel_count; ch++) {
		mixer->channel[ch].in_port = jack_create_in_port( mixer->channel[ch].label, mixer );
	}
	return 0;
}

static int jack_init_out_ports(jmd_mixer_t *mixer)
{
	int ou;
	/* Initialize each of the outputs */
	for(ou=0; ou<mixer->prefader_count+mixer->postfader_count; ou++) {
		/* Create the JACK output ports */
		mixer->output[ou].out_port_l = jack_create_out_port( mixer->output[ou].label, " L" , mixer);
		mixer->output[ou].out_port_r = jack_create_out_port( mixer->output[ou].label, " R", mixer );
	}
	return 0;
}

static int jack_connect_out_port(jmd_mixer_t* mixer, jack_port_t *port, const char* in)
{
	const char* out = jack_port_name(port);
	int err;

	if (verbose) printf("Connecting %s to %s\n", out, in);

	if ((err = jack_connect(mixer->jack_client, out, in)) != 0) {
		fprintf(stderr, "connect_jack_port(): failed to jack_connect() ports: %d\n",err);
		return 1;
	}
	return 0;
}

static int jack_connect_out_ports_auto(jmd_mixer_t* mixer)
{
	/* crude way of automatically connecting up jack ports*/
	const char **all_ports;
	int i;

	/* Get a list of all the jack ports */
	all_ports = jack_get_ports(mixer->jack_client, NULL, NULL, JackPortIsInput);
	if (!all_ports) {
		fprintf(stderr, "autoconnect_jack_ports(): jack_get_ports() returned NULL.");
		return 1;
	}

	/* Step through each port name */
	for (i=0;i<mixer->prefader_count+mixer->postfader_count;i++) {
		/*if (verbose) printf("Port %s\n",all_ports[i]);*/
		/* Connect the port */
		jack_connect_out_port( mixer, mixer->output[i].out_port_l, all_ports[2*i] );
		jack_connect_out_port( mixer, mixer->output[i].out_port_r, all_ports[2*i+1] );
	}
	free(all_ports);
	return 0;
}

static int jack_connect_out_ports_config(jmd_mixer_t* mixer)
{
	int i;

	/* Step through each port name */
	for (i=0;i<mixer->prefader_count+mixer->postfader_count;i++) {
		/* Connect the port */
		if (strcmp(mixer->output[i].to_port_l,"none")){
			jack_connect_out_port( mixer, mixer->output[i].out_port_l,mixer->output[i].to_port_l );
		}
		if (strcmp(mixer->output[i].to_port_r,"none")){
			jack_connect_out_port( mixer, mixer->output[i].out_port_r,mixer->output[i].to_port_r );
		}
	}
	return 0;
}

static int jack_connect_in_port(jmd_mixer_t* mixer, jack_port_t *port, const char* in)
{
	const char* out = jack_port_name(port);
	int err;

	if (verbose) printf("Connecting %s to %s\n", in, out);

	if ((err = jack_connect(mixer->jack_client, in, out)) != 0) {
		fprintf(stderr, "connect_jack_port(): failed to jack_connect() ports: %d\n",err);
		return 1;
	}
	return 0;
}

static int jack_connect_in_ports_auto(jmd_mixer_t* mixer)
{
	/* crude way of automatically connecting up jack ports*/
	const char **all_ports;
	int i;

	/* Get a list of all the jack ports */
	all_ports = jack_get_ports(mixer->jack_client, NULL, NULL, JackPortIsOutput);
	if (!all_ports) {
		fprintf(stderr, "autoconnect_jack_ports(): jack_get_ports() returned NULL.");
		return 1;
	}

	/* Step through each port name */
	for (i=0;i<mixer->channel_count;i++) {
		/*if (verbose) printf("Port %s\n",all_ports[i]);*/
		/* Connect the port */
		jack_connect_in_port( mixer, mixer->channel[i].in_port, all_ports[i] );
	}
	free(all_ports);
	return 0;
}

static int jack_connect_in_ports_config(jmd_mixer_t* mixer)
{
	int i;

	/* Step through each port name */
	for (i=0;i<mixer->channel_count;i++) {
		if (strcmp(mixer->channel[i].from_port,"none")){
			/* Connect the port */
			jack_connect_in_port( mixer, mixer->channel[i].in_port, mixer->channel[i].from_port );
		}
	}
	return 0;
}

static int jack_get_connects(jmd_mixer_t* mixer)
{
	int i;
	for(i=0;i<mixer->channel_count;i++){
		mixer->channel[i].from_ports=NULL;
		if (jack_port_connected(mixer->channel[i].in_port)){
			mixer->channel[i].from_ports=(char **)jack_port_get_connections(mixer->channel[i].in_port);
		}
	}
	for(i=0;i<mixer->prefader_count+mixer->postfader_count;i++){
		mixer->output[i].to_ports_l=NULL;
		if (jack_port_connected(mixer->output[i].out_port_l)){
			mixer->output[i].to_ports_l=(char **)jack_port_get_connections(mixer->output[i].out_port_l);
		}
		mixer->output[i].to_ports_r=NULL;
		if (jack_port_connected(mixer->output[i].out_port_r)){
			mixer->output[i].to_ports_r=(char **)jack_port_get_connections(mixer->output[i].out_port_r);
		}
	}
	return 0;
}

static int lash_init_client(int argc, char *argv[], jmd_mixer_t *mixer)
{
	char lowercase_name[strlen(mixer->client_name)];
	int i;
	lash_event_t *event;

	for (i=0;i<=strlen(mixer->client_name);i++){
		lowercase_name[i]=tolower(mixer->client_name[i]);
	}

	printf("Connecting to LASH...\n");

	mixer->lash_client = lash_init(lash_extract_args(&argc, &argv), lowercase_name,
					   LASH_Config_File,LASH_PROTOCOL(2, 0));

	if (!mixer->lash_client) {
		fprintf(stderr, "%s: could not initialize lash\n", __FUNCTION__);
		return 1;
	}

	event = lash_event_new_with_type(LASH_Client_Name);
	lash_event_set_string(event, mixer->client_name);
	lash_send_event(mixer->lash_client, event);

	printf("Done\n");

	return 0;
}

static int lash_jack_client(jmd_mixer_t *mixer)
{
	lash_event_t *event;

	event = lash_event_new_with_type(LASH_Jack_Client_Name);
	lash_event_set_string(event, mixer->client_name);
	lash_send_event(mixer->lash_client, event);

	return 0;
}

void *lash_process_events(void *ptr){

	lash_event_t *event=NULL;
	jmd_mixer_t *mixer = (jmd_mixer_t*)ptr;
	while(running){
		event = lash_get_event(mixer->lash_client);
		if (event) {
			if (verbose) printf("%s: got event of type %d with string '%s'\n",
				   __FUNCTION__, lash_event_get_type(event),
				   lash_event_get_string(event));
			switch (lash_event_get_type(event)) {
			case LASH_Restore_File:
			{
				sprintf(mixer->xml_config_path,"%s",lash_event_get_string(event));
				lash_send_event(mixer->lash_client, event);
				break;
			}
			case LASH_Save_File:
			{
				sprintf(mixer->xml_config_path,"%s",lash_event_get_string(event));

				/* Retrieve port connections */
				if (verbose) printf("Calling get connections\n");
				if (jack_get_connects(mixer)) {
					fprintf(stderr, "Error getting connections.\n");
					exit(1);
				}
				if (verbose) printf("Finished get connections\n");

				/* Create the config file */
				if (verbose) printf("Calling write config\n");
				if (xml_write_config(mixer)) {
					fprintf(stderr, "Error writing config.\n");
					exit(1);
				}
				if (verbose) printf("Finished write config\n");

				lash_send_event(mixer->lash_client, event);
				break;
			}
			case LASH_Quit:
			{
				running = 0;
				break;
			}
			case LASH_Server_Lost:
			{
				printf("Server connection lost, lashd running?\n");
				break;
			}
			default:
				fprintf(stderr,
						"%s: recieved unknown LASH event of type %d\n",
						__FUNCTION__, lash_event_get_type(event));
				lash_event_destroy(event);
				break;
			}
		}
		sleep(1);
	}
	return 0;
}

static int osc_get_channel_count_handler(const char *path, const char *types, lo_arg **argv, int argc,
		 lo_message msg, void *user_data)
{
	lo_address src = lo_message_get_source( msg );
	jmd_mixer_t *mixer = (jmd_mixer_t*)user_data;
	int result;

	if (verbose) {
    	printf("\n%s received get channel_count OSC message.\n",DEFAULT_CLIENT_NAME);
	}
	/* Send back reply */
	result = lo_send_from( src, mixer->serv, LO_TT_IMMEDIATE, "/mixer/channel_count_is", "i", mixer->channel_count );
	if (result<1) fprintf(stderr, "Error: sending reply failed: %s\n", lo_address_errstr(src));

	return 0;
}

static int osc_get_prefader_count_handler(const char *path, const char *types, lo_arg **argv, int argc,
		 lo_message msg, void *user_data)
{
	lo_address src = lo_message_get_source( msg );
	jmd_mixer_t *mixer = (jmd_mixer_t*)user_data;
	int result;

	if (verbose) {
    	printf("\n%s received get prefader_count OSC message.\n",DEFAULT_CLIENT_NAME);
	}
	/* Send back reply */
	result = lo_send_from( src, mixer->serv, LO_TT_IMMEDIATE, "/mixer/prefader_count_is", "i", mixer->prefader_count );
	if (result<1) fprintf(stderr, "Error: sending reply failed: %s\n", lo_address_errstr(src));

	return 0;
}

static int osc_get_postfader_count_handler(const char *path, const char *types, lo_arg **argv, int argc,
		 lo_message msg, void *user_data)
{
	lo_address src = lo_message_get_source( msg );
	jmd_mixer_t *mixer = (jmd_mixer_t*)user_data;
	int result;

	if (verbose) {
    	printf("\n%s received get postfader_count OSC message.\n",DEFAULT_CLIENT_NAME);
	}
	/* Send back reply */
	result = lo_send_from( src, mixer->serv, LO_TT_IMMEDIATE, "/mixer/postfader_count_is", "i", mixer->postfader_count );
	if (result<1) fprintf(stderr, "Error: sending reply failed: %s\n", lo_address_errstr(src));

	return 0;
}

static int osc_get_output_label_handler(const char *path, const char *types, lo_arg **argv, int argc,
		 lo_message msg, void *user_data)
{
	lo_address src = lo_message_get_source( msg );
	jmd_mixer_t *mixer = (jmd_mixer_t*)user_data;
	int out = argv[0]->i;
	int result;

	if (verbose) {
    	printf("\n%s received output label get OSC message\t",DEFAULT_CLIENT_NAME);
		printf(" (output=%d, label=%s)\n", out, mixer->output[out-1].label);
	}
	/* Send back reply */
	result = lo_send_from( src, mixer->serv, LO_TT_IMMEDIATE, "/mixer/output/label_is", "is", out, mixer->output[out-1].label );
	if (result<1) fprintf(stderr, "Error: sending reply failed: %s\n", lo_address_errstr(src));

	return 0;
}

static int osc_set_output_label_handler(const char *path, const char *types, lo_arg **argv, int argc,
		 lo_message msg, void *user_data)
{
	lo_address src = lo_message_get_source( msg );
	jmd_mixer_t *mixer = (jmd_mixer_t*)user_data;
	int out = argv[0]->i;
	char* label = &argv[1]->s;
	char* port_name_l=NULL;
	char* port_name_r=NULL;
	int result;

	if (verbose) {
    	printf("\n%s received output label change OSC message\t",DEFAULT_CLIENT_NAME);
		printf(" (output=%d, label=%s)\n", out, label);
	}

	if (out < 1 || out > mixer->prefader_count+mixer->postfader_count) {
		fprintf(stderr, "Warning: output number in OSC message is out of range.\n");
		return 1;
	}

	/* store the new value */
	strncpy( mixer->output[out-1].label, label, DEFAULT_LABEL_LEN);
	sprintf(port_name_l,"%s l",label);
	sprintf(port_name_r,"%s r",label);
	jack_port_set_name(mixer->output[out-1].out_port_l,port_name_l);
	jack_port_set_name(mixer->output[out-1].out_port_r,port_name_r);
	/* Send back reply */
	result = lo_send_from( src, mixer->serv, LO_TT_IMMEDIATE, "/mixer/output/label_now", "is", out, mixer->output[out-1].label );
	if (result<1) fprintf(stderr, "Error: sending reply failed: %s\n", lo_address_errstr(src));

	return 0;
}

static int osc_get_output_gain_handler(const char *path, const char *types, lo_arg **argv, int argc,
		 lo_message msg, void *user_data)
{
	lo_address src = lo_message_get_source( msg );
	jmd_mixer_t *mixer = (jmd_mixer_t*)user_data;
	int out = argv[0]->i;
	int result;

	if (verbose) {
    	printf("\n%s received output gain get OSC message\t",DEFAULT_CLIENT_NAME);
		printf(" (ouput=%d, gain=%f)\n", out, mixer->output[out-1].desired_gain);
	}

	/* Send back reply */
	result = lo_send_from( src, mixer->serv, LO_TT_IMMEDIATE, "/mixer/output/gain_is", "if", out, mixer->output[out-1].desired_gain );
	if (result<1) fprintf(stderr, "Error: sending reply failed: %s\n", lo_address_errstr(src));

	return 0;
}

static int osc_set_output_gain_handler(const char *path, const char *types, lo_arg **argv, int argc,
		 lo_message msg, void *user_data)
{
	lo_address src = lo_message_get_source( msg );
	jmd_mixer_t *mixer = (jmd_mixer_t*)user_data;
	int out = argv[0]->i;
	float gain = argv[1]->f;
	int result;

	if (verbose) {
    	printf("\n%s received output gain change OSC message\t",DEFAULT_CLIENT_NAME);
		printf(" (ouput=%d, gain=%fdB)\n", out, gain);
	}

	/* Make sure gain is in range */
	if (gain<-90) gain = -90;
	if (gain>0) gain = 0;

	/* Make sure channel number is in range */
	if (out < 1 || out > mixer->prefader_count+mixer->postfader_count) {
		fprintf(stderr,"Warning: output number in OSC message is out of range.\n");
		return 0;
	}

	/* store the new value */
	mixer->output[out-1].desired_gain = gain;

	if ( mixer->output[out-1].link==1 && out-1 < mixer->prefader_count + mixer->postfader_count){
		mixer->output[out].desired_gain = gain;
	} else if ( mixer->output[out-2].link==1 ){
		mixer->output[out-2].desired_gain = gain;
	}

	/* Send back reply */
	result = lo_send_from( src, mixer->serv, LO_TT_IMMEDIATE, "/mixer/output/gain_now", "if", out, mixer->output[out-1].desired_gain );
	if (result<1) fprintf(stderr, "Error: sending reply failed: %s\n", lo_address_errstr(src));

	return 0;
}

static int osc_get_output_mute_handler(const char *path, const char *types, lo_arg **argv, int argc,
		 lo_message msg, void *user_data)
{
	lo_address src = lo_message_get_source( msg );
	jmd_mixer_t *mixer = (jmd_mixer_t*)user_data;
	int out = argv[0]->i;
	int result;
	int mute_state;

	if (verbose) {
    	printf("\n%s received output mute get OSC message\t",DEFAULT_CLIENT_NAME);
		printf("(output=%d, mute=%f)\n", out, mixer->output[out-1].mute );
	}

	if (out < 1 || out > mixer->prefader_count+mixer->postfader_count) {
		fprintf(stderr, "Warning: output number in OSC message is out of range.\n");
		return 1;
	}

	if (mixer->output[out-1].mute ==0){
		mute_state=0;
	} else {
		mute_state=1;
	}

	/* Send back reply */
	result = lo_send_from( src, mixer->serv, LO_TT_IMMEDIATE, "/mixer/output/mute_is", "ii", out, mute_state);
	if (result<1) fprintf(stderr, "Error: sending reply failed: %s\n", lo_address_errstr(src));

	return 0;
}

static int osc_set_output_mute_handler(const char *path, const char *types, lo_arg **argv, int argc,
		 lo_message msg, void *user_data)
{
	lo_address src = lo_message_get_source( msg );
	jmd_mixer_t *mixer = (jmd_mixer_t*)user_data;
	int out = argv[0]->i;
	int mute = argv[1]->i;
	int result;
	int mute_state;

	if (verbose) {
    	printf("\n%s received output mute set OSC message\t",DEFAULT_CLIENT_NAME);
		printf("(output=%d, mute=%i)\n", out,  mute);
	}

	if (out < 1 || out > mixer->prefader_count+mixer->postfader_count) {
		fprintf(stderr, "Warning: output number in OSC message is out of range.\n");
		return 1;
	}

	/* toggle the value */
	if (mixer->output[out-1].mute ==0){
		mixer->output[out-1].mute = -90;
		mute_state=1;
	} else {
		mixer->output[out-1].mute = 0;
		mute_state=0;
	}

	/* Send back reply */
	result = lo_send_from( src, mixer->serv, LO_TT_IMMEDIATE, "/mixer/output/mute_now", "ii", out, mute_state );
	if (result<1) fprintf(stderr, "Error: sending reply failed: %s\n", lo_address_errstr(src));

	return 0;
}

static int osc_get_output_link_handler(const char *path, const char *types, lo_arg **argv, int argc,
		 lo_message msg, void *user_data)
{
	lo_address src = lo_message_get_source( msg );
	jmd_mixer_t *mixer = (jmd_mixer_t*)user_data;
	int out = argv[0]->i;
	int result;
	int link_state;

	if (verbose) {
    	printf("\n%s received output link get OSC message\t",DEFAULT_CLIENT_NAME);
		printf("(output=%d, link=%d)\n", out, mixer->output[out-1].link );
	}

	if (out < 1 || out > mixer->prefader_count+mixer->postfader_count) {
		fprintf(stderr, "Warning: output number in OSC message is out of range.\n");
		return 1;
	}

	if (mixer->output[out-1].link ==0){
		link_state=0;
	} else {
		link_state=1;
	}

	/* Send back reply */
	result = lo_send_from( src, mixer->serv, LO_TT_IMMEDIATE, "/mixer/output/link_is", "ii", out, link_state);
	if (result<1) fprintf(stderr, "Error: sending reply failed: %s\n", lo_address_errstr(src));

	return 0;
}

static int osc_set_output_link_handler(const char *path, const char *types, lo_arg **argv, int argc,
		 lo_message msg, void *user_data)
{
	lo_address src = lo_message_get_source( msg );
	jmd_mixer_t *mixer = (jmd_mixer_t*)user_data;
	int out = argv[0]->i;
	int mute = argv[1]->i;
	int result;
	int link_state;

	if (verbose) {
    	printf("\n%s received output link set OSC message\t",DEFAULT_CLIENT_NAME);
		printf("(output=%d, link=%d)\n", out,  mute);
	}

	if (out < 1 || out > mixer->prefader_count+mixer->postfader_count) {
		fprintf(stderr, "Warning: output number in OSC message is out of range.\n");
		return 1;
	}

	/* toggle the value */
	if (mixer->output[out-1].link ==0){
		mixer->output[out-1].link = 1;
		link_state=1;
	} else {
		mixer->output[out-1].link = 0;
		link_state=0;
	}
	/* Send back reply */
	result = lo_send_from( src, mixer->serv, LO_TT_IMMEDIATE, "/mixer/output/link_now", "ii", out, link_state );
	if (result<1) fprintf(stderr, "Error: sending reply failed: %s\n", lo_address_errstr(src));

	return 0;
}

static int osc_get_channel_label_handler(const char *path, const char *types, lo_arg **argv, int argc,
		 lo_message msg, void *user_data)
{
	lo_address src = lo_message_get_source( msg );
	jmd_mixer_t *mixer = (jmd_mixer_t*)user_data;
	int chan = argv[0]->i;
	int result;

	if (verbose) {
    	printf("\n%s received channel label get OSC message\t",DEFAULT_CLIENT_NAME);
		printf("(channel=%d, label=%s)\n", chan, mixer->channel[chan-1].label);
	}

	/* Send back reply */
	result = lo_send_from( src, mixer->serv, LO_TT_IMMEDIATE, "/mixer/channel/label_is", "is", chan, mixer->channel[chan-1].label );
	if (result<1) fprintf(stderr, "Error: sending reply failed: %s\n", lo_address_errstr(src));

	return 0;
}

static int osc_set_channel_label_handler(const char *path, const char *types, lo_arg **argv, int argc,
		 lo_message msg, void *user_data)
{
	lo_address src = lo_message_get_source( msg );
	jmd_mixer_t *mixer = (jmd_mixer_t*)user_data;
	int chan = argv[0]->i;
	char* label = &argv[1]->s;
	int result;

	if (verbose) {
    	printf("\n%s received channel label set OSC message\t",DEFAULT_CLIENT_NAME);
		printf("(channel=%d, label=%s)\n", chan, label);
	}

	if (chan < 1 || chan > mixer->channel_count) {
		fprintf(stderr, "Warning: channel number in OSC message is out of range.\n");
		return 1;
	}

	/* store the new value */
	strncpy( mixer->channel[chan-1].label, label, DEFAULT_LABEL_LEN);
	jack_port_set_name(mixer->channel[chan-1].in_port,label);
	/* Send back reply */
	result = lo_send_from( src, mixer->serv, LO_TT_IMMEDIATE, "/mixer/channel/label_now", "is", chan, mixer->channel[chan-1].label );
	if (result<1) fprintf(stderr, "Error: sending reply failed: %s\n", lo_address_errstr(src));

	return 0;
}

static int osc_get_channel_gain_handler(const char *path, const char *types, lo_arg **argv, int argc,
		 lo_message msg, void *user_data)
{
	lo_address src = lo_message_get_source( msg );
	jmd_mixer_t *mixer = (jmd_mixer_t*)user_data;
	int chan = argv[0]->i;
	int send = argv[1]->i;
	int result;

	if (verbose) {
    	printf("\n%s received gain get OSC message\t",DEFAULT_CLIENT_NAME);
		printf("(channel=%d, gain=%f)\n", chan, mixer->channel[chan-1].send[send-1].desired_gain );
	}

	if (chan < 1 || chan > mixer->channel_count) {
		fprintf(stderr, "Warning: channel number in OSC message is out of range.\n");
		return 1;
	}

	if (send < 1 || send > mixer->prefader_count+mixer->postfader_count) {
		fprintf(stderr, "Warning: send number in OSC message is out of range.\n");
		return 1;
	}

	/* Send back reply */
	result = lo_send_from( src, mixer->serv, LO_TT_IMMEDIATE, "/mixer/channel/gain_is", "if", chan, mixer->channel[chan-1].send[send-1].desired_gain );
	if (result<1) fprintf(stderr, "Error: sending reply failed: %s\n", lo_address_errstr(src));

	return 0;
}

static int osc_set_channel_gain_handler(const char *path, const char *types, lo_arg **argv, int argc,
		 lo_message msg, void *user_data)
{
	lo_address src = lo_message_get_source( msg );
	jmd_mixer_t *mixer = (jmd_mixer_t*)user_data;
	int chan = argv[0]->i;
	float gain = argv[1]->f;
	int result;

	/* Make sure gain is in range */
	if (gain<-90) gain = -90;
	if (gain>0) gain = 0; /* thought a channel gain over 0 isn't needed */

	if (verbose) {
    	printf("\n%s received channel gain set OSC message\t",DEFAULT_CLIENT_NAME);
		printf("(channel=%d, gain=%f)\n", chan,  gain);
	}

	if (chan < 1 || chan > mixer->channel_count) {
		fprintf(stderr, "Warning: channel number in OSC message is out of range.\n");
		return 1;
	}

	/* store the new value */
	mixer->channel[chan-1].send[0].desired_gain = gain;

	if ( mixer->channel[chan-1].stereo==1 && chan-1 < mixer->prefader_count + mixer->postfader_count){
		mixer->channel[chan].send[0].desired_gain = gain;
	} else if ( mixer->channel[chan-2].stereo==1 ){
		mixer->channel[chan-2].send[0].desired_gain = gain;
	}
	/* Send back reply */
	result = lo_send_from( src, mixer->serv, LO_TT_IMMEDIATE, "/mixer/channel/gain_now", "if", chan, mixer->channel[chan-1].send[0].desired_gain );
	if (result<1) fprintf(stderr, "Error: sending reply failed: %s\n", lo_address_errstr(src));

	if ( mixer->channel[chan-1].stereo==1 && chan-1 < mixer->channel_count){
		result = lo_send_from( src, mixer->serv, LO_TT_IMMEDIATE, "/mixer/channel/gain_now", "if", chan+1, mixer->channel[chan].send[0].desired_gain );
		if (result<1) fprintf(stderr, "Error: sending reply failed: %s\n", lo_address_errstr(src));
	} else if ( mixer->channel[chan-2].stereo==1 ){
		result = lo_send_from( src, mixer->serv, LO_TT_IMMEDIATE, "/mixer/channel/gain_now", "if", chan-1, mixer->channel[chan-2].send[0].desired_gain );
		if (result<1) fprintf(stderr, "Error: sending reply failed: %s\n", lo_address_errstr(src));
	}


	return 0;
}

static int osc_get_channel_mute_handler(const char *path, const char *types, lo_arg **argv, int argc,
		 lo_message msg, void *user_data)
{
	lo_address src = lo_message_get_source( msg );
	jmd_mixer_t *mixer = (jmd_mixer_t*)user_data;
	int chan = argv[0]->i;
	int result;
	int mute_state;

	if (verbose) {
    	printf("\n%s received channel mute get OSC message\t",DEFAULT_CLIENT_NAME);
		printf("(channel=%d, mute=%f)\n", chan, mixer->channel[chan-1].mute );
	}

	if (chan < 1 || chan > mixer->channel_count) {
		fprintf(stderr, "Warning: channel number in OSC message is out of range.\n");
		return 1;
	}

	if(mixer->channel[chan-1].mute == 0) mute_state=0;
	if(mixer->channel[chan-1].mute == -90) mute_state=1;

	/* Send back reply */
	result = lo_send_from( src, mixer->serv, LO_TT_IMMEDIATE, "/mixer/channel/mute_is", "ii", chan, mute_state );
	if (result<1) fprintf(stderr, "Error: sending reply failed: %s\n", lo_address_errstr(src));

	return 0;
}

static int osc_set_channel_mute_handler(const char *path, const char *types, lo_arg **argv, int argc,
		 lo_message msg, void *user_data)
{
	lo_address src = lo_message_get_source( msg );
	jmd_mixer_t *mixer = (jmd_mixer_t*)user_data;
	int chan = argv[0]->i;
	int mute = argv[1]->i;
	int result;
	int mute_state;

	if (verbose) {
    	printf("\n%s received channel mute set OSC message\t",DEFAULT_CLIENT_NAME);
		printf("(channel=%d, mute=%i)\n", chan,  mute);
	}

	if (chan < 1 || chan > mixer->channel_count) {
		fprintf(stderr, "Warning: channel number in OSC message is out of range.\n");
		return 1;
	}

	/* toggle the value */
	if (mixer->channel[chan-1].mute ==0){
		mixer->channel[chan-1].mute = -90;
		mute_state=1;
	} else {
		mixer->channel[chan-1].mute = 0;
		mute_state=0;
	}

	/* Send back reply */
	result = lo_send_from( src, mixer->serv, LO_TT_IMMEDIATE, "/mixer/channel/mute_now", "ii", chan, mute_state );
	if (result<1) fprintf(stderr, "Error: sending reply failed: %s\n", lo_address_errstr(src));

	return 0;
}

static int osc_get_channel_stereo_handler(const char *path, const char *types, lo_arg **argv, int argc,
		 lo_message msg, void *user_data)
{
	lo_address src = lo_message_get_source( msg );
	jmd_mixer_t *mixer = (jmd_mixer_t*)user_data;
	int chan = argv[0]->i;
	int result;

	if (verbose) {
    	printf("\n%s received stereo get OSC message\t",DEFAULT_CLIENT_NAME);
		printf("(channel=%d, stereo=%i)\n", chan, mixer->channel[chan-1].stereo );
	}

	if (chan < 1 || chan > mixer->channel_count) {
		fprintf(stderr, "Warning: channel number in OSC message is out of range.\n");
		return 1;
	}

	/* Send back reply */
	result = lo_send_from( src, mixer->serv, LO_TT_IMMEDIATE, "/mixer/channel/stereo_is", "ii", chan, mixer->channel[chan-1].stereo );
	if (result<1) fprintf(stderr, "Error: sending reply failed: %s\n", lo_address_errstr(src));

	return 0;
}

static int osc_set_channel_stereo_handler(const char *path, const char *types, lo_arg **argv, int argc,
		 lo_message msg, void *user_data)
{
	lo_address src = lo_message_get_source( msg );
	jmd_mixer_t *mixer = (jmd_mixer_t*)user_data;
	int chan = argv[0]->i;
	int stereo = argv[1]->i;
	int result;

	if (verbose) {
    	printf("\n%s received stereo set OSC message\t",DEFAULT_CLIENT_NAME);
		printf("(channel=%d, stereo=%i)\n", chan,  stereo);
	}

	if (chan < 1 || chan > mixer->channel_count) {
		fprintf(stderr, "Warning: channel number in OSC message is out of range.\n");
		return 1;
	}

	/* toggle the value */
	if (mixer->channel[chan-1].stereo ==0){
		mixer->channel[chan-1].stereo = 1;
	} else {
		mixer->channel[chan-1].stereo = 0;
	}
	/* Send back reply */
	result = lo_send_from( src, mixer->serv, LO_TT_IMMEDIATE, "/mixer/channel/stereo_now", "ii", chan, mixer->channel[chan-1].stereo );
	if (result<1) fprintf(stderr, "Error: sending reply failed: %s\n", lo_address_errstr(src));

	return 0;
}

static int osc_set_channel_pan_handler(const char *path, const char *types, lo_arg **argv, int argc,
		 lo_message msg, void *user_data)
{
	lo_address src = lo_message_get_source( msg );
	jmd_mixer_t *mixer = (jmd_mixer_t*)user_data;
	int chan = argv[0]->i;
	float pan = argv[2]->f;
	int ou;
	int result;

	/* Make sure gain is in range */
	if (pan<-90) pan = -90;
	if (pan>90) pan = 90;

	if (verbose) {
    	printf("\n%s received set pan OSC message\t",DEFAULT_CLIENT_NAME);
		printf("(channel=%d, pan=%f)\n", chan,  pan);
	}

	if (chan < 1 || chan > mixer->channel_count) {
		fprintf(stderr, "Warning: channel number in OSC message is out of range.\n");
		return 1;
	}

	/* store the new value */
	for ( ou=0;ou<mixer->prefader_count+mixer->postfader_count;ou++){
		if ( pan == 0 ){
			mixer->channel[chan-1].send[ou].desired_gain_l = pan;
			mixer->channel[chan-1].send[ou].desired_gain_r = pan;
		} else if ( pan < 0 ){
			mixer->channel[chan-1].send[ou].desired_gain_l = pan;
			mixer->channel[chan-1].send[ou].desired_gain_r = 0;
		} else {
			mixer->channel[chan-1].send[ou].desired_gain_l = 0;
			mixer->channel[chan-1].send[ou].desired_gain_r = pan*(-1);
		}
	}
	/* Send back reply */
	result = lo_send_from( src, mixer->serv, LO_TT_IMMEDIATE, "/mixer/channel/pan_now", "if", chan, pan );
	if (result<1) fprintf(stderr, "Error: sending reply failed: %s\n", lo_address_errstr(src));

	return 0;
}

static int osc_get_channel_send_gain_handler(const char *path, const char *types, lo_arg **argv, int argc,
		 lo_message msg, void *user_data)
{
	lo_address src = lo_message_get_source( msg );
	jmd_mixer_t *mixer = (jmd_mixer_t*)user_data;
	int chan = argv[0]->i;
	int send = argv[1]->i;
	int result;

	if (verbose) {
    	printf("\n%s received send gain get OSC message\t",DEFAULT_CLIENT_NAME);
		printf("(channel=%d, send=%d, gain %f)\n", chan, send, mixer->channel[chan-1].send[send-1].desired_gain);
	}

	if (chan < 1 || chan > mixer->channel_count) {
		fprintf(stderr, "Warning: channel number in OSC message is out of range.\n");
		return 1;
	}

	if (send < 1 || send > mixer->prefader_count+mixer->postfader_count) {
		fprintf(stderr, "Warning: send number in OSC message is out of range.\n");
		return 1;
	}
	/* Send back reply */
	result = lo_send_from( src, mixer->serv, LO_TT_IMMEDIATE, "/mixer/channel/send/gain_is", "iif", chan, send, mixer->channel[chan-1].send[send-1].desired_gain );
	if (result<1) fprintf(stderr, "Error: sending reply failed: %s\n", lo_address_errstr(src));

	return 0;
}

static int osc_set_channel_send_gain_handler(const char *path, const char *types, lo_arg **argv, int argc,
		 lo_message msg, void *user_data)
{
	lo_address src = lo_message_get_source( msg );
	jmd_mixer_t *mixer = (jmd_mixer_t*)user_data;
	int chan = argv[0]->i;
	int send = argv[1]->i;
	float gain = argv[2]->f;
	int result;

	/* Make sure gain is in range */
	if (gain<-90) gain = -90;
	if (gain>0) gain = 0; /* thought a channel gain over 0 isn't needed */

	if (verbose) {
    	printf("\n%s received send gain set OSC message\t",DEFAULT_CLIENT_NAME);
		printf("(channel=%d, send=%d, gain=%f)\n", chan, send, gain);
	}

	if (chan < 1 || chan > mixer->channel_count) {
		fprintf(stderr, "Warning: channel number in OSC message is out of range.\n");
		return 1;
	}

	if (send < 1 || send > mixer->prefader_count+mixer->postfader_count) {
		fprintf(stderr, "Warning: send number in OSC message is out of range.\n");
		return 1;
	}

	/* store the new value */
	mixer->channel[chan-1].send[send-1].desired_gain = gain;

	if ( mixer->channel[chan-1].stereo==1 && chan-1 < mixer->prefader_count + mixer->postfader_count){
		mixer->channel[chan].send[send-1].desired_gain = gain;
	}
	if ( mixer->channel[chan-2].stereo==1 ){
		mixer->channel[chan-2].send[send-1].desired_gain = gain;
	}

	/* Send back reply */
	result = lo_send_from( src, mixer->serv, LO_TT_IMMEDIATE, "/mixer/channel/send/gain_now", "iif", chan, send, mixer->channel[chan-1].send[send-1].desired_gain );
	if (result<1) fprintf(stderr, "Error: sending reply failed: %s\n", lo_address_errstr(src));

	if ( mixer->channel[chan-1].stereo==1 && chan-1 < mixer->channel_count){
		result = lo_send_from( src, mixer->serv, LO_TT_IMMEDIATE, "/mixer/channel/send/gain_now", "iif", chan+1, send, mixer->channel[chan].send[send-1].desired_gain );
		if (result<1) fprintf(stderr, "Error: sending reply failed: %s\n", lo_address_errstr(src));
	} else if ( mixer->channel[chan-2].stereo==1 ){
		result = lo_send_from( src, mixer->serv, LO_TT_IMMEDIATE, "/mixer/channel/send/gain_now", "iif", chan-1, send, mixer->channel[chan-2].send[send-1].desired_gain );
		if (result<1) fprintf(stderr, "Error: sending reply failed: %s\n", lo_address_errstr(src));
	}

	return 0;
}

static int osc_get_channel_send_pan_handler(const char *path, const char *types, lo_arg **argv, int argc,
		 lo_message msg, void *user_data)
{
	lo_address src = lo_message_get_source( msg );
	jmd_mixer_t *mixer = (jmd_mixer_t*)user_data;
	int chan = argv[0]->i;
	int send = argv[1]->i;
	float pan;
	int result;

	if (chan < 1 || chan > mixer->channel_count) {
		fprintf(stderr, "Warning: channel number in OSC message is out of range.\n");
		return 1;
	}

	if (send < 1 || send > mixer->prefader_count+mixer->postfader_count) {
		fprintf(stderr, "Warning: send number in OSC message is out of range.\n");
		return 1;
	}

	if ( mixer->channel[chan-1].send[send-1].desired_gain_l == 0 && mixer->channel[chan-1].send[send-1].desired_gain_r == 0 ){
		pan=0;
	} else if ( mixer->channel[chan-1].send[send-1].desired_gain_l == 0 && mixer->channel[chan-1].send[send-1].desired_gain_r < 0 ){
		pan=mixer->channel[chan-1].send[send-1].desired_gain_r*(-1);
	} else {
		pan=mixer->channel[chan-1].send[send-1].desired_gain_l;
	}

	if (verbose) {
    	printf("\n%s received send get pan OSC message\t",DEFAULT_CLIENT_NAME);
		printf("(channel=%d, send=%d, pan=%f)\n", chan, send, pan);
	}

	/* Send back reply */
	result = lo_send_from( src, mixer->serv, LO_TT_IMMEDIATE, "/mixer/channel/send/pan_is", "iif", chan, send, pan );
	if (result<1) fprintf(stderr, "Error: sending reply failed: %s\n", lo_address_errstr(src));

	return 0;
}

static int osc_set_channel_send_pan_handler(const char *path, const char *types, lo_arg **argv, int argc,
		 lo_message msg, void *user_data)
{
	lo_address src = lo_message_get_source( msg );
	jmd_mixer_t *mixer = (jmd_mixer_t*)user_data;
	int chan = argv[0]->i;
	int send = argv[1]->i;
	float pan = argv[2]->f;
	int result;

	/* Make sure gain is in range */
	if (pan<-90) pan = -90;
	if (pan>90) pan = 90;

	if (verbose) {
    	printf("\n%s received send set pan OSC message\t",DEFAULT_CLIENT_NAME);
		printf("(channel=%d, send=%d, pan=%f)\n", chan, send, pan);
	}

	if (chan < 1 || chan > mixer->channel_count) {
		fprintf(stderr, "Warning: channel number in OSC message is out of range.\n");
		return 1;
	}

	if (send < 1 || send > mixer->prefader_count+mixer->postfader_count) {
		fprintf(stderr, "Warning: send number in OSC message is out of range.\n");
		return 1;
	}

	/* store the new value */
	if ( pan == 0 ){
		mixer->channel[chan-1].send[send-1].desired_gain_l = pan;
		mixer->channel[chan-1].send[send-1].desired_gain_r = pan;
	} else if ( pan < 0 ){
		mixer->channel[chan-1].send[send-1].desired_gain_l = pan;
		mixer->channel[chan-1].send[send-1].desired_gain_r = 0;
	} else {
		mixer->channel[chan-1].send[send-1].desired_gain_l = 0;
		mixer->channel[chan-1].send[send-1].desired_gain_r = pan*(-1);
	}
	/* Send back reply */
	result = lo_send_from( src, mixer->serv, LO_TT_IMMEDIATE, "/mixer/channel/send/pan_now", "iif", chan, send, pan );
	if (result<1) fprintf(stderr, "Error: sending reply failed: %s\n", lo_address_errstr(src));

	return 0;
}

static int osc_get_channel_send_gain_l_handler(const char *path, const char *types, lo_arg **argv, int argc,
		 lo_message msg, void *user_data)
{
	lo_address src = lo_message_get_source( msg );
	jmd_mixer_t *mixer = (jmd_mixer_t*)user_data;
	int chan = argv[0]->i;
	int send = argv[1]->i;
	int result;

	if (verbose) {
    	printf("\n%s received send gain get left OSC message\t",DEFAULT_CLIENT_NAME);
		printf("(channel=%d, send=%d, gain=%f)\n", chan, send, mixer->channel[chan-1].send[send-1].desired_gain_l);
	}

	if (chan < 1 || chan > mixer->channel_count) {
		fprintf(stderr, "Warning: channel number in OSC message is out of range.\n");
		return 1;
	}

	if (send < 1 || send > mixer->prefader_count+mixer->postfader_count) {
		fprintf(stderr, "Warning: send number in OSC message is out of range.\n");
		return 1;
	}
		/* Send back reply */
	result = lo_send_from( src, mixer->serv, LO_TT_IMMEDIATE, "/mixer/channel/send/gain_l_is", "iif", chan, send, mixer->channel[chan-1].send[send-1].desired_gain_l );
	if (result<1) fprintf(stderr, "Error: sending reply failed: %s\n", lo_address_errstr(src));

	return 0;
}

static int osc_set_channel_send_gain_l_handler(const char *path, const char *types, lo_arg **argv, int argc,
		 lo_message msg, void *user_data)
{
	lo_address src = lo_message_get_source( msg );
	jmd_mixer_t *mixer = (jmd_mixer_t*)user_data;
	int chan = argv[0]->i;
	int send = argv[1]->i;
	float gain = argv[2]->f;
	int result;

	/* Make sure gain is in range */
	if (gain<-90) gain = -90;
	if (gain>0) gain = 0; /* thought a channel gain over 0 isn't needed */

	if (verbose) {
    	printf("\n%s received send gain set left OSC message\t",DEFAULT_CLIENT_NAME);
		printf("(channel=%d, send=%d, gain=%f)\n", chan, send, gain);
	}

	if (chan < 1 || chan > mixer->channel_count) {
		fprintf(stderr, "Warning: channel number in OSC message is out of range.\n");
		return 1;
	}

	if (send < 1 || send > mixer->prefader_count+mixer->postfader_count) {
		fprintf(stderr, "Warning: send number in OSC message is out of range.\n");
		return 1;
	}

	/* store the new value
	 */
	mixer->channel[chan-1].send[send-1].desired_gain_l = gain;

	/* Send back reply */
	result = lo_send_from( src, mixer->serv, LO_TT_IMMEDIATE, "/mixer/channel/send/gain_l_now", "iif", chan, send, mixer->channel[chan-1].send[send-1].desired_gain_l );
	if (result<1) fprintf(stderr, "Error: sending reply failed: %s\n", lo_address_errstr(src));

	return 0;
}

static int osc_get_channel_send_gain_r_handler(const char *path, const char *types, lo_arg **argv, int argc,
		 lo_message msg, void *user_data)
{
	lo_address src = lo_message_get_source( msg );
	jmd_mixer_t *mixer = (jmd_mixer_t*)user_data;
	int chan = argv[0]->i;
	int send = argv[1]->i;
	int result;

	if (verbose) {
    	printf("\n%s received send gain get right OSC message\t",DEFAULT_CLIENT_NAME);
		printf("(channel=%d, send=%d, gain=%f)\n", chan, send, mixer->channel[chan-1].send[send-1].desired_gain_r);
	}

	if (chan < 1 || chan > mixer->channel_count) {
		fprintf(stderr, "Warning: channel number in OSC message is out of range.\n");
		return 1;
	}

	if (send < 1 || send > mixer->prefader_count+mixer->postfader_count) {
		fprintf(stderr, "Warning: send number in OSC message is out of range.\n");
		return 1;
	}

	/* Send back reply */
	result = lo_send_from( src, mixer->serv, LO_TT_IMMEDIATE, "/mixer/channel/send/gain_r_is", "iif", chan, send, mixer->channel[chan-1].send[send-1].desired_gain_r );
	if (result<1) fprintf(stderr, "Error: sending reply failed: %s\n", lo_address_errstr(src));

	return 0;
}

static int osc_set_channel_send_gain_r_handler(const char *path, const char *types, lo_arg **argv, int argc,
		 lo_message msg, void *user_data)
{
	lo_address src = lo_message_get_source( msg );
	jmd_mixer_t *mixer = (jmd_mixer_t*)user_data;
	int chan = argv[0]->i;
	int send = argv[1]->i;
	float gain = argv[2]->f;
	int result;

	/* Make sure gain is in range */
	if (gain<-90) gain = -90;
	if (gain>0) gain = 0; /* thought a channel gain over 0 isn't needed */

	if (verbose) {
    	printf("\n%s received send gain set right OSC message\t",DEFAULT_CLIENT_NAME);
		printf("(channel=%d, send=%d, gain=%f)\n", chan, send, gain);
	}

	if (chan < 1 || chan > mixer->channel_count) {
		fprintf(stderr, "Warning: channel number in OSC message is out of range.\n");
		return 1;
	}

	if (send < 1 || send > mixer->prefader_count+mixer->postfader_count) {
		fprintf(stderr, "Warning: send number in OSC message is out of range.\n");
		return 1;
	}

	/* store the new value */
	mixer->channel[chan-1].send[send-1].desired_gain_r = gain;

	/* Send back reply */
	result = lo_send_from( src, mixer->serv, LO_TT_IMMEDIATE, "/mixer/channel/send/gain_r_now", "iif", chan, send, mixer->channel[chan-1].send[send-1].desired_gain_r );
	if (result<1) fprintf(stderr, "Error: sending reply failed: %s\n", lo_address_errstr(src));

	return 0;
}

static int osc_save_handler(const char *path, const char *types, lo_arg **argv, int argc,
		 lo_message msg, void *user_data)
{
	lo_address src = lo_message_get_source( msg );
	jmd_mixer_t *mixer = (jmd_mixer_t*)user_data;
	sprintf(mixer->xml_config_path,"%s",&argv[0]->s);
	sprintf(mixer->xml_config_file,"%s",&argv[1]->s);
	int result;

	if (jack_get_connects(mixer)) {
		fprintf(stderr, "Error getting connections.\n");
		return 1;
	}
	/* Create the config file */
	if (xml_write_config(mixer)) {
		fprintf(stderr, "Error writing config.\n");
		return 1;
	}

	/* Send back reply */
	result = lo_send_from( src, mixer->serv, LO_TT_IMMEDIATE, "/mixer/saved", "ss", mixer->xml_config_path, mixer->xml_config_file );
	if (result<1) fprintf(stderr, "Error: sending reply failed: %s\n", lo_address_errstr(src));

	return 0;
}

static int osc_quit_handler(const char *path, const char *types, lo_arg **argv, int argc,
		 lo_message msg, void *user_data)
{
	lo_address src = lo_message_get_source( msg );
	jmd_mixer_t *mixer = (jmd_mixer_t*)user_data;
	int result;

    printf("\nquitting\n\n");

	/* Send back reply */
	result = lo_send_from( src, mixer->serv, LO_TT_IMMEDIATE, "/exit", "" );
	if (result<1) fprintf(stderr, "Error: sending reply failed: %s\n", lo_address_errstr(src));

    running = 0;
    fflush(stdout);

    return 0;
}

static int osc_ping_handler(const char *path, const char *types, lo_arg **argv, int argc,
		 lo_message msg, void *user_data)
{
	lo_address src = lo_message_get_source( msg );
	jmd_mixer_t *mixer = (jmd_mixer_t*)user_data;
	int result;

	/* Display the address the ping came from */
	if (verbose) {
		char *url = lo_address_get_url(src);
		printf( "\nGot ping from: %s\n", url);
		free(url);
	}

	/* Send back reply */
	result = lo_send_from( src, mixer->serv, LO_TT_IMMEDIATE, "/pong", "" );
	if (result<1) fprintf(stderr, "Error: sending reply failed: %s\n", lo_address_errstr(src));

    return 0;
}

static int osc_generic_handler(const char *path, const char *types, lo_arg **argv,
		    int argc, void *data, void *user_data)
{
    int i;

	if(!quiet){
		printf("\n%s received unknown OSC message.\n",DEFAULT_CLIENT_NAME);
	    printf("path: <%s>\n", path);
	    for (i=0; i<argc; i++) {
			printf("arg %d '%c' ", i, types[i]);
			lo_arg_pp(types[i], argv[i]);
			printf("\n");
	    }
	}
    fflush(stdout);
    return 1;
}

static void osc_server_error_handler(int num, const char *msg, const char *path)
{
    fprintf(stderr, "\nLibLO server error %d in path %s: %s\n", num, path, msg);
    fflush(stdout);
}

static int osc_server_init(jmd_mixer_t *mixer)
{
	/* Create new server */
	mixer->st = lo_server_thread_new( mixer->osc_mixer_port, osc_server_error_handler );

	mixer->serv = lo_server_thread_get_server( mixer->st );

	/* Add the methods */
    lo_server_thread_add_method(mixer->st, "/mixer/channel_count_get", "", osc_get_channel_count_handler, mixer);
    lo_server_thread_add_method(mixer->st, "/mixer/prefader_count_get", "", osc_get_prefader_count_handler, mixer);
    lo_server_thread_add_method(mixer->st, "/mixer/postfader_count_get", "", osc_get_postfader_count_handler, mixer);

    lo_server_thread_add_method(mixer->st, "/mixer/output/label_get", "i", osc_get_output_label_handler, mixer);
    lo_server_thread_add_method(mixer->st, "/mixer/output/label_set", "is", osc_set_output_label_handler, mixer);
    lo_server_thread_add_method(mixer->st, "/mixer/output/gain_get", "i", osc_get_output_gain_handler, mixer);
    lo_server_thread_add_method(mixer->st, "/mixer/output/gain_set", "if", osc_set_output_gain_handler, mixer);
    lo_server_thread_add_method(mixer->st, "/mixer/output/mute_get", "i", osc_get_output_mute_handler, mixer);
    lo_server_thread_add_method(mixer->st, "/mixer/output/mute_set", "ii", osc_set_output_mute_handler, mixer);
    lo_server_thread_add_method(mixer->st, "/mixer/output/link_get", "i", osc_get_output_link_handler, mixer);
    lo_server_thread_add_method(mixer->st, "/mixer/output/link_set", "ii", osc_set_output_link_handler, mixer);

    lo_server_thread_add_method(mixer->st, "/mixer/channel/label_get", "i", osc_get_channel_label_handler, mixer);
    lo_server_thread_add_method(mixer->st, "/mixer/channel/label_set", "is", osc_set_channel_label_handler, mixer);
    lo_server_thread_add_method(mixer->st, "/mixer/channel/gain_get", "i", osc_get_channel_gain_handler, mixer);
    lo_server_thread_add_method(mixer->st, "/mixer/channel/gain_set", "if", osc_set_channel_gain_handler, mixer);
    lo_server_thread_add_method(mixer->st, "/mixer/channel/pan_set", "if", osc_set_channel_pan_handler, mixer);
    lo_server_thread_add_method(mixer->st, "/mixer/channel/mute_get", "i", osc_get_channel_mute_handler, mixer);
    lo_server_thread_add_method(mixer->st, "/mixer/channel/mute_set", "ii", osc_set_channel_mute_handler, mixer);
    lo_server_thread_add_method(mixer->st, "/mixer/channel/stereo_get", "i", osc_get_channel_stereo_handler, mixer);
    lo_server_thread_add_method(mixer->st, "/mixer/channel/stereo_set", "ii", osc_set_channel_stereo_handler, mixer);

    lo_server_thread_add_method(mixer->st, "/mixer/channel/send/gain_get", "ii", osc_get_channel_send_gain_handler, mixer);
    lo_server_thread_add_method(mixer->st, "/mixer/channel/send/gain_set", "iif", osc_set_channel_send_gain_handler, mixer);
    lo_server_thread_add_method(mixer->st, "/mixer/channel/send/pan_get", "ii", osc_get_channel_send_pan_handler, mixer);
    lo_server_thread_add_method(mixer->st, "/mixer/channel/send/pan_set", "iif", osc_set_channel_send_pan_handler, mixer);
    lo_server_thread_add_method(mixer->st, "/mixer/channel/send/gain_l_get", "ii", osc_get_channel_send_gain_l_handler, mixer);
    lo_server_thread_add_method(mixer->st, "/mixer/channel/send/gain_l_set", "iif", osc_set_channel_send_gain_l_handler, mixer);
    lo_server_thread_add_method(mixer->st, "/mixer/channel/send/gain_r_get", "ii", osc_get_channel_send_gain_r_handler, mixer);
    lo_server_thread_add_method(mixer->st, "/mixer/channel/send/gain_r_set", "iif", osc_set_channel_send_gain_r_handler, mixer);
    lo_server_thread_add_method(mixer->st, "/save", "ss", osc_save_handler, mixer);
	lo_server_thread_add_method(mixer->st, "/ping", "", osc_ping_handler, mixer);
    lo_server_thread_add_method(mixer->st, "/quit", "", osc_quit_handler, mixer);

    /* add method that will match any path and args, got to be last */
    lo_server_thread_add_method(mixer->st, NULL, NULL, osc_generic_handler, mixer);

	/* Start the thread */
	lo_server_thread_start(mixer->st);

	if (!quiet) {
		char *url = lo_server_thread_get_url(mixer->st);
		printf("OSC server URL: %s\n",url);
		free(url);
	}
	return 0;
}

static void osc_server_finish(jmd_mixer_t *mixer)
{
	lo_server_thread_stop( mixer->st );
	lo_server_thread_free( mixer->st );

	if (verbose) printf( "Stopped OSC server thread.\n");
}

jmd_mixer_t* jmd_init_mixer()
{
	jmd_mixer_t *mixer  = malloc(sizeof(jmd_mixer_t));
	mixer->autoconnect = 0;
	mixer->channel_count = DEFAULT_CHANNEL_COUNT;
	mixer->prefader_count = DEFAULT_PREFADER_COUNT;
	mixer->postfader_count = DEFAULT_POSTFADER_COUNT;
	sprintf(mixer->osc_mixer_port,"%s",DEFAULT_OSC_PORT);
	sprintf(mixer->client_name,"%s",DEFAULT_CLIENT_NAME);
	sprintf(mixer->xml_config_file,"%s",DEFAULT_CONFIG_FILE);
	sprintf(mixer->xml_config_path,"%s",DEFAULT_CONFIG_PATH);
	mixer->jack_client = NULL;
	mixer->channel = NULL;
	mixer->output = NULL;

	return(mixer);
}

static void jmd_signal_handler(int signum)
{
	if (!quiet) {
		switch(signum) {
			case SIGTERM:	fprintf(stderr, "Got termination signal.\n"); break;
			case SIGINT:	fprintf(stderr, "Got interupt signal.\n"); break;
		}
	}
	running=0;
}

static int jmd_init_chan_struct(jmd_mixer_t *mixer)
{
	int ch;

	mixer->channel = (jmd_channel_t*)malloc(
			sizeof(jmd_channel_t) * mixer->channel_count);

	/* Initialise each of the channels */
	for(ch=0; ch<mixer->channel_count; ch++) {

		mixer->channel[ch].send = (jmd_send_t*)malloc(
			sizeof(jmd_send_t) * (mixer->prefader_count+mixer->postfader_count));
	}
	return 0;
}

static int jmd_init_out_struct(jmd_mixer_t *mixer)
{
	mixer->output = (jmd_output_t*)malloc(
			sizeof(jmd_output_t) * (mixer->prefader_count+mixer->postfader_count));

	return 0;
}

static int jmd_init_chan_data(jmd_mixer_t *mixer)
{
	int ch;
	int s;
	/* Initialise each of the channels data */
	for(ch=0; ch<mixer->channel_count; ch++) {
		snprintf( mixer->channel[ch].label, DEFAULT_LABEL_LEN, "Channel %3d", ch+1 );
		sprintf( mixer->channel[ch].from_port, "none" );
		mixer->channel[ch].from_ports=NULL;
		mixer->channel[ch].stereo=0;
		mixer->channel[ch].mute=0;

		/* Initialise each of the sends */
		for(s=0; s<mixer->prefader_count+mixer->postfader_count;s++) {

			if (s==0) { /* if even channel send[0] left to master with 0dB */
				mixer->channel[ch].send[s].current_gain=0.0f;
				mixer->channel[ch].send[s].desired_gain=0.0f;
				mixer->channel[ch].send[s].current_gain_l=0.0f;
				mixer->channel[ch].send[s].desired_gain_l=0.0f;
				mixer->channel[ch].send[s].current_gain_r=0.0f;
				mixer->channel[ch].send[s].desired_gain_r=0.0f;
			} else {	/* channel to other outs -90dB  */
				mixer->channel[ch].send[s].current_gain=-90.0f;
				mixer->channel[ch].send[s].desired_gain=-90.0f;
				mixer->channel[ch].send[s].current_gain_l=0.0f;
				mixer->channel[ch].send[s].desired_gain_l=0.0f;
				mixer->channel[ch].send[s].current_gain_r=0.0f;
				mixer->channel[ch].send[s].desired_gain_r=0.0f;
			}
			if (verbose==2) {
				printf("channel %i send %i gain %f gain_l %f gain_r %f label %s\n",ch,s,mixer->channel[ch].send[s].current_gain,mixer->channel[ch].send[s].current_gain_l,mixer->channel[ch].send[s].current_gain_r,mixer->output[s].label);
			}
		}
	}
	return 0;
}

static int jmd_init_out_data(jmd_mixer_t *mixer)
{
	char name[10];
	int ou;
	int pr=1;
	int po=1;

	/* Initialize each of the outputs */
	for(ou=0; ou<mixer->prefader_count+mixer->postfader_count; ou++) {
		if ( ou==0 ) {
			sprintf( name, "Master" );
			pr=pr+1;
		} else if ( ou==1 && ou<mixer->prefader_count) {
			sprintf( name, "Rear" );
			pr=pr+1;
		} else if ( ou>1 && ou<mixer->prefader_count) {
			sprintf( name, "Pre %2d", pr );
			pr=pr+1;
		} else  {
			sprintf( name, "Post %2d", po );
			po=po+1;
		}

		snprintf( mixer->output[ou].label, DEFAULT_LABEL_LEN, "%s",name );
		mixer->output[ou].to_ports_l=NULL;
		mixer->output[ou].to_ports_r=NULL;
		sprintf(mixer->output[ou].to_port_l,"none");
		sprintf(mixer->output[ou].to_port_r,"none");

		mixer->output[ou].mute = 0;
		mixer->output[ou].link = 0;

		if ( ou==0 ) { /* Master out with half gain */
			mixer->output[ou].current_gain=-45.0f;
			mixer->output[ou].desired_gain=-45.0f;
		} else  { /* Other sends muted */
			mixer->output[ou].current_gain=-90.0f;
			mixer->output[ou].desired_gain=-90.0f;
		}
	}
	return 0;
}

static int jmd_label_ports(int argc, char *argv[], jmd_mixer_t *mixer)
{
	int i=0;
	/* Label the channels */
	for(i=0; i<argc && i<mixer->channel_count; i++) {
		strncpy( mixer->channel[i].label, argv[i], DEFAULT_LABEL_LEN );
	}
	/* Label the outputs */
	for(; i<argc && i<(mixer->channel_count+mixer->prefader_count+mixer->postfader_count); i++) {
		strncpy( mixer->output[i-mixer->channel_count].label, argv[i], DEFAULT_LABEL_LEN );
	}
	return 0;
}

void *jmd_process_gains(void *ptr)
{
	int c,s;
	jmd_mixer_t *mixer = (jmd_mixer_t*)ptr;

	double old_gain;

	while (running) {
		/* fade output gains */
		for(s=0;s<mixer->prefader_count+mixer->postfader_count;s++){
				mixer->output[s].current_gain=mixer->output[s].desired_gain;
		}

		/* Step through the channels */
		for (c=0;c<mixer->channel_count;c++){

			for(s=0;s<mixer->prefader_count+mixer->postfader_count;s++){
				/* fade send gains */
					mixer->channel[c].send[s].current_gain=mixer->channel[c].send[s].desired_gain;
				/* fade send gains left*/
					mixer->channel[c].send[s].current_gain_l=mixer->channel[c].send[s].desired_gain_l;

				/* fade send gains right */
					mixer->channel[c].send[s].current_gain_r=mixer->channel[c].send[s].desired_gain_r;

				/* calculate mix gains */
				if(s<mixer->prefader_count){
					/* prefader gains */
					old_gain = mixer->channel[c].send[s].mix_gain_l;
					mixer->channel[c].send[s].mix_gain_l=
						db2lin(mixer->channel[c].mute+mixer->channel[c].send[s].current_gain+mixer->channel[c].send[s].current_gain_l+mixer->output[s].current_gain+mixer->output[s].mute);
					mixer->channel[c].send[s].mix_gain_r=
						db2lin(mixer->channel[c].mute+mixer->channel[c].send[s].current_gain+mixer->channel[c].send[s].current_gain_r+mixer->output[s].current_gain+mixer->output[s].mute);
				} else {
					/* postfader gains */
					old_gain = mixer->channel[c].send[s].mix_gain_r;
					mixer->channel[c].send[s].mix_gain_l=
						db2lin(mixer->channel[c].mute+mixer->channel[c].send[0].current_gain+mixer->channel[c].send[s].current_gain+mixer->channel[c].send[s].current_gain_l+mixer->output[s].current_gain+mixer->output[s].mute);
					mixer->channel[c].send[s].mix_gain_r=
						db2lin(mixer->channel[c].mute+mixer->channel[c].send[0].current_gain+mixer->channel[c].send[s].current_gain+mixer->channel[c].send[s].current_gain_r+mixer->output[s].current_gain+mixer->output[s].mute);
				}
			}
		}
		usleep(GAIN_FADE_SLEEP);
	}
	return 0;
}

static void jmd_finish_channels(jmd_mixer_t* mixer)
{
	free(mixer->channel);
}

static void jmd_finish_outputs(jmd_mixer_t* mixer)
{
	free(mixer->output);
}

static void jmd_finish_mixer(jmd_mixer_t* mixer)
{
	free(mixer);
}

static int jmd_usage()
{
	printf("Copyright (C) 2007 Nicholas J Humfrey <njh@ecs.soton.ac.uk>\n");
	printf("                   Ulrich Lorenz Schlueter <audio-mobster@gmx.de>\n\n");
	printf("This program comes with ABSOLUTELY NO WARRANTY.  You are licensed to use it\n");
	printf("under the terms of the GNU General Public License, version 2 or later.  See\n");
	printf("the COPYING file that came with this software for details.\n\n");
	printf("JackMixDesk version %s\n\n", PACKAGE_VERSION);
	printf("Usage: %s [options] [<channel label> ...]\n", PACKAGE_NAME);
	printf("   -a            Autoconnect ports\n");
	printf("   -c <count>    Number of input channels (default 4)\n");
	printf("   -e <count>    Number of pre fader sends (default 0)\n");
	printf("   -o <count>    Number of post fader sends (default 0)\n");
	printf("   -m <port>     Set the OSC UDP port number default '16000'\n");
	printf("   -n <name>     Name for this JACK client (default jackmixdesk)\n");
	printf("   -f <file>     Name for config file (default .mixdeskrc)\n");
	printf("   -p <path>     Path for config file (default $HOME)\n");
	printf("   -v            Enable verbose mode\n");
	printf("   -q            Enable quiet mode\n");
	printf("\n");
	exit(1);
}

static void jmd_welcome(jmd_mixer_t* mixer)
{
	/* Dislay welcoming message */
	if (verbose) {
		printf("JackMixDesk OSC Port: %s\n",mixer->osc_mixer_port);
		printf("Starting JackMixDesk version %s with %d channels %d pre and %d post sends.\n",VERSION, mixer->channel_count, mixer->prefader_count, mixer->postfader_count );
		printf("First prefader send is to master and the gain is used for post sends.\n");
		printf("Second prefader send is called rear.\n");
	}

}

int main(int argc, char *argv[])
{
	jmd_mixer_t *mixer  = NULL;
	pthread_t gain_thread;
	pthread_t lash_thread;
	int opt;
	int opt_ind=0;		/* getopt_long option index */

	/* Set signal handlers */
	signal(SIGTERM, jmd_signal_handler);
	signal(SIGINT, jmd_signal_handler);

    /*
     * this initializes the xml2 library and checks potential ABI mismatches
     * between the version it was compiled for and the actual shared
     * library used.
     */
	LIBXML_TEST_VERSION

	mixer=jmd_init_mixer();

	/* TEST Parse base of config file, overridden by commandline args */
	if (xml_parse_base_config(mixer)) {
		fprintf(stderr, "No valid config file found, we'll create one.\n");
	}

	/* Parse the command line arguments overrides config file */
	while ((opt = getopt_long(argc, argv, "avqhc:e:o:n:m:f:p:l:s:i:",opts_long, &opt_ind)) != -1) {
		switch (opt) {
			case 'a':  mixer->autoconnect=1; break;
			case 'v':  verbose++; break;
			case 'q':  quiet++; break;
			case 'c':  mixer->channel_count = atoi(optarg); break;
			case 'e':  mixer->prefader_count = atoi(optarg); break;
			case 'o':  mixer->postfader_count = atoi(optarg); break;
			case 'n':  sprintf(mixer->client_name,"%s",optarg); break;
			case 'm':  snprintf(mixer->osc_mixer_port,strlen(optarg),"%s",optarg); break;
			case 'f':  sprintf(mixer->xml_config_file,"%s",optarg); break;
			case 'p':  sprintf(mixer->xml_config_path,"%s",optarg); break;
			case 'l':  sprintf(mixer->lash_project,"%s",optarg); break;
			case 's':  sprintf(mixer->lash_server,"%s",optarg); break;
			case 'i':  sprintf(mixer->lash_id,"%s",optarg); break;
			default:
				fprintf(stderr, "Unknown option '%c'.\n", (char)opt);
			case 'h':
				jmd_usage(argv[0]); break;
		}
	}

	/* Setup LASH client */
	if (!lash_init_client(argc,argv,mixer)) {
		/* Start lash event thread */
		pthread_create( &lash_thread, NULL,lash_process_events , mixer );
	} else {
		printf("LASH support disabled.\n");
	}

	/* Wait for LASH_Restore_File event to be processed */
	sleep(3); /* mutex on lash_process_events? */

	/* LASH needs arguments unchanged */
	argc -= optind;
    argv += optind;

	/* Check parameters */
	if (mixer->channel_count<1) jmd_usage();
	if (mixer->prefader_count<1) jmd_usage();	/* Needed for master out */

	/* Display welcome message */
	jmd_welcome(mixer);

	/* Setup JACK client */
	if (jack_init_client(mixer)) {
		fprintf(stderr, "Error initializing jack client.\n");
		exit(1);
	}

	/* Send Jack client name to LASH */
	if (lash_jack_client(mixer)) {
		fprintf(stderr, "Error sending Jack name to LASH.\n");
		exit(1);
	}

	/* malloc the output structure */
	if (jmd_init_out_struct(mixer)) {
		fprintf(stderr, "Error initializing output data.\n");
		exit(1);
	}

	/* malloc the channel structure */
	if (jmd_init_chan_struct(mixer)) {
		fprintf(stderr, "Error initializing channel data.\n");
		exit(1);
	}

	/* Create the output descriptors */
	if (jmd_init_out_data(mixer)) {
		fprintf(stderr, "Error initializing output data.\n");
		exit(1);
	}

	/* Create the channel descriptors */
	if (jmd_init_chan_data(mixer)) {
		fprintf(stderr, "Error initializing channel data.\n");
		exit(1);
	}

	/* Parse base of config file, will be overridden by commandline args */
	if (xml_parse_io_config(mixer)) {
		fprintf(stderr, "No valid config file found.\n");
	}

	/* Label the channels and then outputs with passed args */
	if (jmd_label_ports(argc,argv,mixer)) {
		fprintf(stderr, "Error labeling ports.\n");
		exit(1);
	}

	/* Create jack output ports */
	if (jack_init_out_ports(mixer)) {
		fprintf(stderr, "Error initializing out ports.\n");
		exit(1);
	}

	/* Create jack channel ports */
	if (jack_init_chan_ports(mixer)) {
		fprintf(stderr, "Error initializing in ports.\n");
		exit(1);
	}

	/* Set JACK running */
	if (jack_activate(mixer->jack_client)) {
		fprintf(stderr, "Error activating jack client.\n");
		exit(1);
	}

	/* Setup OSC */
	if (osc_server_init(mixer)) {
		fprintf(stderr, "Error initializing OSC server thread.\n");
		exit(1);
	}

	if (mixer->autoconnect){
		/* autoconnect in ports */
		if (jack_connect_in_ports_auto(mixer)) {
			fprintf(stderr, "Error auto connecting in ports.\n");
		}
		/* autoconnect out ports */
		if (jack_connect_out_ports_auto(mixer)) {
			fprintf(stderr, "Error auto connecting out ports.\n");
		}
	} else {
		/* connect in ports */
		if (jack_connect_in_ports_config(mixer)) {
			fprintf(stderr, "Error connecting in ports.\n");
		}
		/* connect out ports */
		if (jack_connect_out_ports_config(mixer)) {
			fprintf(stderr, "Error connecting out ports.\n");
		}
	}

	/* Start gain processing thread */
	pthread_create( &gain_thread, NULL, jmd_process_gains, mixer );

	/* Sleep until we are done (work is done in threads) */
	while (running) {
		sleep(1);
	}

	/* Get the connects and write xml if no lash server */
	if(!lash_server_connected(mixer->lash_client))
	{
		if (verbose==2) printf("Calling get connections\n");
		if (jack_get_connects(mixer)) {
			fprintf(stderr, "Error getting connections.\n");
			exit(1);
		}
		if (verbose==2) printf("Finished get connections\n");

		/* Create the config file */
		if (verbose==2) printf("Calling write config\n");
		if (xml_write_config(mixer)) {
			fprintf(stderr, "Error writing config.\n");
			exit(1);
		} else {
    		xmlFreeDoc(mixer->doc);
		}
		if (verbose==2) printf("Finished write config\n");
	}

	/* Shutdown OSC server */
	osc_server_finish(mixer);

	/* Shutdown jack client */
	jack_finish_client(mixer);

	/* Cleanup */
	jmd_finish_channels(mixer);
	jmd_finish_outputs(mixer);
	jmd_finish_mixer(mixer);

	printf("Quitting JackMixDesk.\n");

	return 0;
}
